from __future__ import annotations

import abc
import inspect
import sys
import uuid
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, cast

import executing
import pyarrow as pa

from chalk._expression.converter import (
    convert_literal_to_proto_expr,
    convert_pa_dtype_to_proto_expr,
    convert_proto_expr_to_literal,
)
from chalk._gen.chalk.expression.v1 import expression_pb2 as expr_pb2

if TYPE_CHECKING:
    from chalk.features._encoding.converter import TPrimitive
from chalk.utils.environment_parsing import env_var_bool
from chalk.utils.source_parsing import should_skip_source_code_parsing

SUPPORTED_UNDERSCORE_OPS_BINARY = set("+ - * / // % ** < <= > >= == != & | ^ << >>".split())
SUPPORTED_UNDERSCORE_OPS_UNARY = set("- + ~".split())

if sys.version_info < (3, 11):
    # On python < 3.11, this can cause a significant perf regression, so we always skip parsing
    CHALK_SKIP_PRECISE_UNDERSCORE_SOURCE_COLUMN = True
else:
    CHALK_SKIP_PRECISE_UNDERSCORE_SOURCE_COLUMN = env_var_bool("CHALK_SKIP_PRECISE_UNDERSCORE_SOURCE_COLUMN")
    """
    If this environment variable is set, then `Underscore` expressions will not attempt to record their
    location within a class.

    For feature classes with thousands of members, this can add up to several seconds of validation time.
    However, if this is enabled, then error messages cannot include accurate line/column numbers for
    underscore expressions.
    """


@dataclass
class UnderscoreDefinitionLocation:
    file: str
    line: int
    column: int


def convert_value_to_proto_expr(value: Any) -> expr_pb2.LogicalExprNode:
    if isinstance(value, Underscore):
        return value._to_proto()  # pyright: ignore[reportPrivateUsage]

    return convert_literal_to_proto_expr(value)


class Underscore:
    """An unevaluated underscore expression.

    Examples
    --------
    >>> class X:
    ...     y: DataFrame[Y] = has_many(...)
    ...     s: int = _.y[_.z].sum()
    """

    def __init__(self, expr_id: Optional[str] = None):
        super().__init__()
        current_frame = inspect.currentframe()
        while current_frame is not None and current_frame.f_code.co_filename.endswith("/chalk/features/underscore.py"):
            current_frame = current_frame.f_back
        if current_frame is not None:
            definition_location = UnderscoreDefinitionLocation(
                file=current_frame.f_code.co_filename,
                line=current_frame.f_lineno,
                column=1,
                # Currently, this is just a guess, due to lack of reliable column info directly on Python frame
            )
            if not should_skip_source_code_parsing() and not CHALK_SKIP_PRECISE_UNDERSCORE_SOURCE_COLUMN:
                # Attempt to get the exact AST node to get a more-precise caller location
                # for error reporting purposes:
                source_node = executing.Source.executing(current_frame).node
                if hasattr(source_node, "lineno") and source_node.lineno is not None:  # pyright: ignore
                    definition_location.line = source_node.lineno  # pyright: ignore
                if hasattr(source_node, "col_offset") and source_node.col_offset is not None:  # pyright: ignore
                    definition_location.column = source_node.col_offset  # pyright: ignore
            del current_frame
        else:
            definition_location = None

        self._chalk_definition_location: Optional[UnderscoreDefinitionLocation] = definition_location
        self._chalk__expr_id = expr_id if expr_id is not None else str(uuid.uuid4())

    def definition_location(self) -> Optional[UnderscoreDefinitionLocation]:
        return self._chalk_definition_location

    def __getattr__(self, attr: str) -> "Underscore":
        if attr.startswith("__") or attr.startswith("_chalk__"):
            raise AttributeError(f"{self.__class__.__name__!r} {attr!r}")
        return UnderscoreAttr(self, attr)

    def __getitem__(self, key: Any) -> "Underscore":
        return UnderscoreItem(self, key)

    def __call__(self, *args: Any, **kwargs: Any) -> "Underscore":
        return UnderscoreCall(self, *args, **kwargs)

    def __add__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("+", self, other)

    def __radd__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("+", other, self)

    def __sub__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("-", self, other)

    def __rsub__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("-", other, self)

    def __mul__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("*", self, other)

    def __rmul__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("*", other, self)

    def __truediv__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("/", self, other)

    def __rtruediv__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("/", other, self)

    def __floordiv__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("//", self, other)

    def __rfloordiv__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("//", other, self)

    def __mod__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("%", self, other)

    def __rmod__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("%", other, self)

    def __pow__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("**", self, other)

    def __rpow__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("**", other, self)

    def __lt__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("<", self, other)

    def __le__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("<=", self, other)

    def __gt__(self, other: Any) -> "Underscore":
        return UnderscoreFunction(">", self, other)

    def __ge__(self, other: Any) -> "Underscore":
        return UnderscoreFunction(">=", self, other)

    def __eq__(self, other: Any) -> "Underscore":  # pyright: ignore[reportIncompatibleMethodOverride]
        return UnderscoreFunction("==", self, other)

    def __ne__(self, other: Any) -> "Underscore":  # pyright: ignore[reportIncompatibleMethodOverride]
        return UnderscoreFunction("!=", self, other)

    def __and__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("&", self, other)

    def __rand__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("&", other, self)

    def __or__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("|", self, other)

    def __ror__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("|", other, self)

    def __xor__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("^", self, other)

    def __rxor__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("^", other, self)

    def __lshift__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("<<", self, other)

    def __rlshift__(self, other: Any) -> "Underscore":
        return UnderscoreFunction("<<", other, self)

    def __rshift__(self, other: Any) -> "Underscore":
        return UnderscoreFunction(">>", self, other)

    def __rrshift__(self, other: Any) -> "Underscore":
        return UnderscoreFunction(">>", other, self)

    def __neg__(self) -> "Underscore":
        return UnderscoreFunction("negate", self)

    def __pos__(self) -> "Underscore":
        # Returning self as "+" by itself really doesn't make sense
        return self

    def __invert__(self) -> "Underscore":
        return UnderscoreFunction("~", self)

    def __hash__(self):
        # Underscores may not be hashable, so hashing the type instead. This probably won't cause a perf bomb because it's unlikely to have a hash table of underscore expressions
        return hash(type(self))

    @abc.abstractmethod
    def _is_equal(self, other: Underscore) -> bool:
        """Return whether this underscore is equal to another underscore. Normally == would provide this, but we override it to allow
        composite building over underscore expressions"""
        raise NotImplementedError

    @abc.abstractmethod
    def _to_proto(self) -> expr_pb2.LogicalExprNode:
        raise NotImplementedError

    @classmethod
    def _from_proto(cls, node: expr_pb2.LogicalExprNode) -> Underscore | TPrimitive | pa.Scalar:
        value = cls._from_proto_dispatch(node)
        if isinstance(value, Underscore):
            if node.expr_id:
                value._chalk__expr_id = node.expr_id
        return value

    @classmethod
    def _from_proto_dispatch(cls, node: expr_pb2.LogicalExprNode) -> Underscore | TPrimitive | pa.Scalar:
        if node.HasField("identifier"):
            if node.identifier.name == "_":
                return UnderscoreRoot()
            elif node.identifier.name == "__":
                return DoubleUnderscore()
            else:
                raise ValueError(f"Unknown identifier: {node.identifier.name}")
        elif node.HasField("get_attribute"):
            parent = cls._from_proto(node.get_attribute.parent)
            if not isinstance(parent, Underscore):
                raise TypeError(f"Parent is not an underscore. Got {parent}")
            return UnderscoreAttr(
                parent=parent,
                attr=node.get_attribute.attribute.name,
            )
        elif node.HasField("get_subscript"):
            parent = cls._from_proto(node.get_subscript.parent)
            if not isinstance(parent, Underscore):
                raise TypeError(f"Parent is not an underscore. Got {parent}")
            subscripts = tuple(cls._from_proto(sub) for sub in node.get_subscript.subscript)
            if len(subscripts) > 0:
                return UnderscoreItem(parent=parent, key=subscripts)
            else:
                raise ValueError("empty subscript")
        elif node.HasField("call"):
            args = node.call.args
            if not node.call.func.HasField("identifier"):
                func = cls._from_proto(node.call.func)
                if not isinstance(func, Underscore):
                    raise TypeError(f"Func is not an underscore. Got {func}")
                return UnderscoreCall(
                    func,
                    *[cls._from_proto(arg) for arg in args],
                    _chalk__expr_id=node.expr_id,
                    **{k: cls._from_proto(v) for k, v in node.call.kwargs.items()},
                )
            func_name = node.call.func.identifier.name

            if func_name == "cast":
                # Cast is special, because it needs to have the output dtype
                if len(args) != 2:
                    raise ValueError("cast should have exactly two args -- the parent, and a pseudoarg for the dtype")
                parent = cls._from_proto(args[0])
                if not isinstance(parent, Underscore):
                    raise ValueError(f"The parent for cast should be an underscore; got {parent}")
                dest_dummy_scalar = cls._from_proto(args[1])
                if not isinstance(dest_dummy_scalar, pa.Scalar):
                    raise ValueError(f"The pseudoarg for cast should be a pyarrow scalar; got {dest_dummy_scalar}")
                return UnderscoreCast(parent, cast(pa.Scalar, dest_dummy_scalar).type)
            return UnderscoreFunction(
                func_name,
                *[cls._from_proto(arg) for arg in args],
                _chalk__expr_id=node.expr_id,
                **{k: cls._from_proto(v) for k, v in node.call.kwargs.items()},
            )
        elif node.HasField("literal_value"):
            return convert_proto_expr_to_literal(node)
        else:
            raise ValueError(f"Unknown LogicalExprNode type for underscores: {node.WhichOneof('expr_form')}")

    def __iter__(self):
        raise TypeError(f"{self.__class__.__name__} object is not iterable")


class UnderscoreRoot(Underscore):
    # _
    def __repr__(self):
        return "_"

    def _to_proto(self) -> expr_pb2.LogicalExprNode:
        return expr_pb2.LogicalExprNode(identifier=expr_pb2.Identifier(name="_"), expr_id=self._chalk__expr_id)

    def _is_equal(self, other: Underscore):
        return isinstance(other, UnderscoreRoot)  # All root underscores are equal to each other

    @property
    def chalk_window(self):
        """Refers to the specific window being evaluated in the context of
        defining an underscore expression.

        Examples
        --------
        >>> from datetime import timedelta
        >>> from chalk.features import features
        >>> from chalk import _, Windowed, DataFrame, windowed
        >>> @features
        ... class Transaction:
        ...     id: int
        ...     user_id: "User.id"
        ...     amount: int
        >>> @features
        ... class User:
        ...     id: int
        ...     transactions: DataFrame[Transaction]
        ...     sum_amounts: Windowed[int] = windowed(
        ...         "30d", "90d",
        ...         expression=_.transactions[
        ...             _.amount,
        ...             _.ts > _.chalk_window
        ...         ].sum(),
        ...     )
        """
        return UnderscoreAttr(self, "chalk_window")

    @property
    def chalk_now(self):
        """Refers to the specific window being evaluated in the context of
        defining an underscore expression.

        Examples
        --------
        >>> from datetime import timedelta
        >>> from chalk.features import features
        >>> from chalk import _, Windowed, DataFrame, windowed
        >>> @features
        ... class Transaction:
        ...     id: int
        ...     user_id: "User.id"
        ...     amount: int
        >>> @features
        ... class User:
        ...     id: int
        ...     transactions: DataFrame[Transaction]
        ...     sum_old_amounts: int = _.transactions[
        ...         _.amount,
        ...         _.ts < _.chalk_now - timedelta(days=30),
        ...     ].sum()
        """
        return UnderscoreAttr(self, "chalk_now")

    def if_then_else(
        self,
        condition: Underscore,
        if_true: Any,
        if_false: Any,
    ) -> Underscore:
        """
        Create a conditional expression, roughly equivalent to

        ```
        if condition:
            return if_true
        else:
            return if_false
        ```

        The input `condition` is always evaluated first. The `if_true` and `if_false` expressions
        are only evaluated if the condition is true or false, respectively.

        Examples
        --------
        >>> from chalk.features import features, _
        >>> @features
        ... class Transaction:
        ...    id: int
        ...    amount: int
        ...    risk_score: bool = _.if_then_else(
        ...        _.amount > 10_000,
        ...        _.amount * 0.1,
        ...        _.amount * 0.05,
        ...    )
        """
        return UnderscoreFunction("if_else", condition, if_true, if_false)


class DoubleUnderscore(Underscore):
    # __
    def __repr__(self):
        return "__"

    def _is_equal(self, other: Underscore) -> bool:
        return isinstance(other, DoubleUnderscore)  # all double underscores are equal to each other

    def _to_proto(self) -> expr_pb2.LogicalExprNode:
        return expr_pb2.LogicalExprNode(identifier=expr_pb2.Identifier(name="__"))


class UnderscoreAttr(Underscore):
    # _.a
    def __init__(self, parent: Underscore, attr: str, *, expr_id: Optional[str] = None):
        super().__init__(expr_id)
        self._chalk__parent = parent
        self._chalk__attr = attr

    def __repr__(self):
        return f"{self._chalk__parent}.{self._chalk__attr}"

    def _is_equal(self, other: Underscore) -> bool:
        if self is other:
            return True
        if not isinstance(other, UnderscoreAttr):
            return False
        return self._chalk__parent._is_equal(other._chalk__parent) and self._chalk__attr == other._chalk__attr

    def _to_proto(self) -> expr_pb2.LogicalExprNode:
        return expr_pb2.LogicalExprNode(
            expr_id=self._chalk__expr_id,
            get_attribute=expr_pb2.ExprGetAttribute(
                parent=convert_value_to_proto_expr(self._chalk__parent),
                attribute=expr_pb2.Identifier(name=self._chalk__attr),
            ),
        )


def _are_args_equal(x: object, y: object):
    if isinstance(x, Underscore):
        if not isinstance(y, Underscore):
            return False
        return x._is_equal(y)  # pyright: ignore[reportPrivateUsage]
    else:
        return x == y


class UnderscoreItem(Underscore):
    # _[k]
    def __init__(self, parent: Underscore, key: Any, *, expr_id: Optional[str] = None):
        super().__init__(expr_id)
        self._chalk__parent = parent
        self._chalk__key = key if isinstance(key, tuple) else (key,)

    def __repr__(self):
        keys = ", ".join(f"{key}" for key in self._chalk__key)
        return f"{self._chalk__parent}[{keys}]"

    def _is_equal(self, other: Underscore) -> bool:
        if self is other:
            return True
        if not isinstance(other, UnderscoreItem):
            return False
        if not self._chalk__parent._is_equal(other._chalk__parent):
            return False
        if len(self._chalk__key) != len(other._chalk__key):
            return False
        for x, y in zip(self._chalk__key, other._chalk__key):
            if not _are_args_equal(x, y):
                return False
        return True

    def _to_proto(self) -> expr_pb2.LogicalExprNode:
        raw_key = self._chalk__key
        converted_keys: list[expr_pb2.LogicalExprNode] = []
        converted_keys.extend(convert_value_to_proto_expr(k) for k in raw_key)

        return expr_pb2.LogicalExprNode(
            expr_id=self._chalk__expr_id,
            get_subscript=expr_pb2.ExprGetSubscript(
                parent=convert_value_to_proto_expr(self._chalk__parent),
                subscript=converted_keys,
            ),
        )


class UnderscoreCall(Underscore):
    # _(args, kwargs)
    def __init__(self, parent: Underscore, *args: Any, _chalk__expr_id: Optional[str] = None, **kwargs: Any):
        super().__init__(_chalk__expr_id)
        self._chalk__parent = parent
        self._chalk__args = args
        self._chalk__kwargs = kwargs

    def __repr__(self):
        args: list[str] = []
        COMMA = ", "
        for arg in self._chalk__args:
            args.append(f"{arg}")
        for key, arg in self._chalk__kwargs.items():
            args.append(f"{key}={arg}")
        return f"{self._chalk__parent}({COMMA.join(args)})"

    def _to_proto(self) -> expr_pb2.LogicalExprNode:
        args = [convert_value_to_proto_expr(arg) for arg in self._chalk__args]
        kwargs = {k: convert_value_to_proto_expr(v) for k, v in self._chalk__kwargs.items()}
        return expr_pb2.LogicalExprNode(
            expr_id=self._chalk__expr_id,
            call=expr_pb2.ExprCall(
                func=convert_value_to_proto_expr(self._chalk__parent),
                args=args,
                kwargs=kwargs,
            ),
        )

    def _is_equal(self, other: Underscore) -> bool:
        if self is other:
            return True
        if not isinstance(other, UnderscoreCall):
            return False
        if not self._chalk__parent._is_equal(other._chalk__parent):
            return False

        if len(self._chalk__args) != len(other._chalk__args):
            return False
        for x, y in zip(self._chalk__args, other._chalk__args):
            if not _are_args_equal(x, y):
                return False

        if len(self._chalk__kwargs) != len(other._chalk__kwargs):
            return False
        for k, x in self._chalk__kwargs.items():
            if k not in other._chalk__kwargs:
                return False
            y = other._chalk__kwargs[k]

            if not _are_args_equal(x, y):
                return False
        return True


def coerce_dtype_args_to_scalar(arg: Any) -> Any:
    """
    Currently Chalk underscore gRPC can only transport scalars, so we need to convert any dtype arguments to a pa.scalar.
    """

    if isinstance(arg, pa.DataType):
        return pa.scalar(None, arg)
    return arg


class UnderscoreFunction(Underscore):
    __name__ = "function"

    def __init__(self, name: str, *args: Any, _chalk__expr_id: Optional[str] = None, **kwargs: Any):
        super().__init__(_chalk__expr_id)
        self._chalk__args = args
        self._chalk__kwargs = kwargs
        self._chalk__args = [coerce_dtype_args_to_scalar(arg) for arg in args]

        self._chalk__function_name = name

    def __repr__(self):
        if self._chalk__function_name in SUPPORTED_UNDERSCORE_OPS_UNARY and len(self._chalk__args) == 1:
            return f"({self._chalk__function_name} {self._chalk__args[0]!r})"
        if self._chalk__function_name in SUPPORTED_UNDERSCORE_OPS_BINARY and len(self._chalk__args) == 2:
            return f"({self._chalk__args[0]!r} {self._chalk__function_name} {self._chalk__args[1]!r})"
        else:
            fn: list[str] = [f"_.{self._chalk__function_name}("]
            fn.append(", ".join(repr(x) for x in self._chalk__args))
            if self._chalk__args and self._chalk__kwargs:
                fn.append(", ")
            fn.append(", ".join(f"{k}={v!r}" for k, v in self._chalk__kwargs.items()))
            fn.append(")")
            return "".join(fn)

    def _to_proto(self) -> expr_pb2.LogicalExprNode:
        return expr_pb2.LogicalExprNode(
            expr_id=self._chalk__expr_id,
            call=expr_pb2.ExprCall(
                func=expr_pb2.LogicalExprNode(identifier=expr_pb2.Identifier(name=self._chalk__function_name)),
                args=[convert_value_to_proto_expr(x) for x in self._chalk__args],
                kwargs={k: convert_value_to_proto_expr(v) for (k, v) in self._chalk__kwargs.items()},
            ),
        )

    def _is_equal(self, other: Underscore) -> bool:
        if self is other:
            return True
        if not isinstance(other, UnderscoreFunction):
            return False
        if self._chalk__function_name != other._chalk__function_name:
            return False

        if len(self._chalk__args) != len(other._chalk__args):
            return False
        for x, y in zip(self._chalk__args, other._chalk__args):
            if not _are_args_equal(x, y):
                return False

        if len(self._chalk__kwargs) != len(other._chalk__kwargs):
            return False
        for k, x in self._chalk__kwargs.items():
            if k not in other._chalk__kwargs:
                return False
            y = other._chalk__kwargs[k]

            if not _are_args_equal(x, y):
                return False
        return True

    def __bool__(self):
        # because __eq__ is overridden, `if underscore_1 == underscore_2: ...` would always return True
        # To fix that, we'll manually override bool for the case when the expression is an UnderscoreFunction and the operator is == or !=
        if self._chalk__function_name == "==":
            lhs = self._chalk__args[0]
            rhs = self._chalk__args[1]
            return _are_args_equal(lhs, rhs)
        if self._chalk__function_name == "!=":
            lhs = self._chalk__args[0]
            rhs = self._chalk__args[1]
            return not _are_args_equal(lhs, rhs)
        # Ideally, we would raise, but this causes issues when using the ast-parsing syntax like
        # cheap_and_good_tacos = df[_.price <= 12 and _.rating >= 4.0]
        # Retuning True to match the default behavior of python's __bool__
        return True


class UnderscoreCast(Underscore):
    __name__ = "cast"
    __qualname__ = "chalk.functions.cast"

    def __init__(self, value: Underscore, to_type: pa.DataType, *, expr_id: Optional[str] = None):
        super().__init__(expr_id)
        self._chalk__value = value
        self._chalk__to_type = to_type

    def _is_equal(self, other: Underscore) -> bool:
        if self is other:
            return True
        if not isinstance(other, UnderscoreCast):
            return False
        if not self._chalk__value._is_equal(other._chalk__value):
            return False
        if self._chalk__to_type != other._chalk__to_type:
            return False
        return True

    def __repr__(self):
        return f"chalk.functions.cast({self._chalk__value}, {self._chalk__to_type})"

    def _to_proto(self) -> expr_pb2.LogicalExprNode:
        return expr_pb2.LogicalExprNode(
            expr_id=self._chalk__expr_id,
            call=expr_pb2.ExprCall(
                func=expr_pb2.LogicalExprNode(identifier=expr_pb2.Identifier(name="cast")),
                args=[
                    convert_value_to_proto_expr(self._chalk__value),
                    convert_pa_dtype_to_proto_expr(self._chalk__to_type),
                ],
            ),
        )


_ = underscore = UnderscoreRoot()
__ = DoubleUnderscore()

# NEED `__all__` because `_` is private and can't be auto-imported by i.e. IntelliJ.
__all__ = (
    "SUPPORTED_UNDERSCORE_OPS_BINARY",
    "SUPPORTED_UNDERSCORE_OPS_UNARY",
    "Underscore",
    "UnderscoreAttr",
    "UnderscoreCall",
    "UnderscoreCast",
    "UnderscoreFunction",
    "UnderscoreItem",
    "UnderscoreRoot",
    "_",
    "__",
    "underscore",
)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ChalkKubernetesEventSeries(_message.Message):
    __slots__ = ("count", "last_observed_time")
    COUNT_FIELD_NUMBER: _ClassVar[int]
    LAST_OBSERVED_TIME_FIELD_NUMBER: _ClassVar[int]
    count: int
    last_observed_time: int
    def __init__(self, count: _Optional[int] = ..., last_observed_time: _Optional[int] = ...) -> None: ...

class ChalkKubernetesEvent(_message.Message):
    __slots__ = ("type", "reason", "action", "reporting_controller", "event_time", "note", "series")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    ACTION_FIELD_NUMBER: _ClassVar[int]
    REPORTING_CONTROLLER_FIELD_NUMBER: _ClassVar[int]
    EVENT_TIME_FIELD_NUMBER: _ClassVar[int]
    NOTE_FIELD_NUMBER: _ClassVar[int]
    SERIES_FIELD_NUMBER: _ClassVar[int]
    type: str
    reason: str
    action: str
    reporting_controller: str
    event_time: int
    note: str
    series: ChalkKubernetesEventSeries
    def __init__(
        self,
        type: _Optional[str] = ...,
        reason: _Optional[str] = ...,
        action: _Optional[str] = ...,
        reporting_controller: _Optional[str] = ...,
        event_time: _Optional[int] = ...,
        note: _Optional[str] = ...,
        series: _Optional[_Union[ChalkKubernetesEventSeries, _Mapping]] = ...,
    ) -> None: ...



from datetime import datetime
from typing import List

from chalk.features.feature_field import DUMMY_FEATURE, Feature

__all__ = [
    "CHALK_TS_FEATURE",
    "Now",
    "DUMMY_FEATURE",
    "ID_FEATURE",
    "OBSERVED_AT_FEATURE",
    "REPLACED_OBSERVED_AT_FEATURE",
    "Distance",
    "PSEUDOFEATURES",
    "FQN_OR_NAME_TO_PSEUDOFEATURE",
    "FQN_TO_PSEUDOFEATURE",
    "PSEUDONAMESPACE",
]

PSEUDONAMESPACE = "__chalk__"

CHALK_TS_FEATURE = Feature(
    name="CHALK_TS",
    namespace=PSEUDONAMESPACE,
    typ=datetime,
    max_staleness=None,
    etl_offline_to_online=False,
    primary=False,
    is_feature_time=False,
    is_pseudofeature=True,
)

Now = Feature(
    name="now",
    namespace=PSEUDONAMESPACE,
    typ=datetime,
    max_staleness=None,
    etl_offline_to_online=False,
    primary=False,
    is_feature_time=False,
    is_pseudofeature=True,
)

ID_FEATURE = Feature(
    name="__id__",
    namespace=PSEUDONAMESPACE,
    typ=str,
    max_staleness=None,
    etl_offline_to_online=False,
    primary=False,
    is_feature_time=False,
    is_pseudofeature=True,
)

OBSERVED_AT_FEATURE = Feature(
    name="__observed_at__",
    namespace=PSEUDONAMESPACE,
    typ=datetime,
    max_staleness=None,
    etl_offline_to_online=False,
    primary=False,
    is_feature_time=False,
    is_pseudofeature=True,
)

REPLACED_OBSERVED_AT_FEATURE = Feature(
    name="__replaced_observed_at__",
    namespace=PSEUDONAMESPACE,
    typ=datetime,
    max_staleness=None,
    etl_offline_to_online=False,
    primary=False,
    is_feature_time=False,
    is_pseudofeature=True,
)

Distance = Feature(
    name="__distance__",
    namespace=PSEUDONAMESPACE,
    typ=float,
    max_staleness=None,
    etl_offline_to_online=False,
    primary=False,
    is_feature_time=False,
    is_pseudofeature=True,
)

PSEUDOFEATURES: List[Feature] = [
    CHALK_TS_FEATURE,
    DUMMY_FEATURE,
    Distance,
    ID_FEATURE,
    Now,
    OBSERVED_AT_FEATURE,
    REPLACED_OBSERVED_AT_FEATURE,
]

FQN_TO_PSEUDOFEATURE = {f.fqn: f for f in PSEUDOFEATURES}
FQN_OR_NAME_TO_PSEUDOFEATURE = dict(**FQN_TO_PSEUDOFEATURE, **{f.name: f for f in PSEUDOFEATURES})



def _can_use_ddtrace() -> bool:
    try:
        from ddtrace import __version__ as version

        # Parse version string into components (e.g. "2.6.0" -> [2, 6, 0])
        version_parts = version.split(".")

        # Check if it's version 2.6.0 or higher
        if len(version_parts) >= 2:
            major = int(version_parts[0])
            minor = int(version_parts[1])

            # Allow ddtrace 2.6+ to be used
            return major == 2 and minor >= 6
        return False
    except Exception:
        return False


def _can_use_datadog_statsd() -> bool:
    try:
        from datadog.dogstatsd.base import statsd

        _ = statsd
        return True
    except ImportError:
        return False


can_use_ddtrace = _can_use_ddtrace()
can_use_datadog_statsd = _can_use_datadog_statsd()



from __future__ import annotations

from enum import Enum
from typing import Any, List

import pandas as pd

from chalk.features import DataFrame

MAX_REPR_DEPTH = 2


def get_repr_value(raw: Any, depth: int = 0) -> object:
    limit = 10 if depth < MAX_REPR_DEPTH else 0
    if isinstance(raw, (pd.DataFrame, DataFrame)):
        return f"DataFrame[shape={raw.shape}]"
    elif isinstance(raw, Enum):
        return raw.value
    elif isinstance(raw, dict) and "columns" in raw and "values" in raw and len(raw) == 2:
        # If feature not in registry, has-many results will not
        # be converted into a DataFrame, bringing us to this block.
        cols, vals = raw["columns"], raw["values"]
        num_cols = num_rows = "unknown no. of"
        if isinstance(cols, list):
            num_cols = len(cols)
        if isinstance(vals, list) and vals and isinstance(vals[0], list):
            num_rows = len(vals[0])
        col_tag = "col" if num_cols == 1 else "cols"
        row_tag = "row" if num_rows == 1 else "rows"
        return f"has_many[ {num_rows} {row_tag} x {num_cols} {col_tag}  ]"
    elif isinstance(raw, list):
        reprs: List[str] = []
        for elem in raw[:limit]:
            if isinstance(elem, (dict, list)):
                elem_repr = get_repr_value(elem, depth + 1)
                if type(elem_repr) is not str:
                    # Must convert the element to a `str` so that we can
                    # concatenate later.
                    elem_repr = repr(elem_repr)
                reprs.append(elem_repr)
            else:
                reprs.append(repr(elem))
        if len(raw) > limit:
            reprs.append("...")

        list_contents = ", ".join(reprs)
        return f"[{list_contents}]"
    elif isinstance(raw, dict):
        reprs: List[str] = []
        idx = 0
        for key, val in raw.items():
            if idx >= limit:
                break
            idx += 1
            key_repr = repr(key)
            if isinstance(val, (dict, list)):
                value_repr = get_repr_value(val, depth + 1)
            else:
                value_repr = repr(val)
            reprs.append(f"{key_repr}: {value_repr}")
        if len(raw) > limit:
            reprs.append("...")
        dict_contents = ", ".join(reprs)
        return f"{{{dict_contents}}}"
    else:
        return raw



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/auth/v1/featurepermission.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n%chalk/auth/v1/featurepermission.proto\x12\rchalk.auth.v1"\xb0\x01\n\x12\x46\x65\x61turePermissions\x12?\n\x04tags\x18\x01 \x03(\x0b\x32+.chalk.auth.v1.FeaturePermissions.TagsEntryR\x04tags\x1aY\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x36\n\x05value\x18\x02 \x01(\x0e\x32 .chalk.auth.v1.FeaturePermissionR\x05value:\x02\x38\x01*\x99\x01\n\x11\x46\x65\x61turePermission\x12"\n\x1e\x46\x45\x41TURE_PERMISSION_UNSPECIFIED\x10\x00\x12\x1c\n\x18\x46\x45\x41TURE_PERMISSION_ALLOW\x10\x01\x12%\n!FEATURE_PERMISSION_ALLOW_INTERNAL\x10\x02\x12\x1b\n\x17\x46\x45\x41TURE_PERMISSION_DENY\x10\x03\x42\x91\x01\n\x11\x63om.chalk.auth.v1B\x16\x46\x65\x61turepermissionProtoP\x01Z\x0e\x61uth/v1;authv1\xa2\x02\x03\x43\x41X\xaa\x02\rChalk.Auth.V1\xca\x02\rChalk\\Auth\\V1\xe2\x02\x19\x43halk\\Auth\\V1\\GPBMetadata\xea\x02\x0f\x43halk::Auth::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.auth.v1.featurepermission_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\021com.chalk.auth.v1B\026FeaturepermissionProtoP\001Z\016auth/v1;authv1\242\002\003CAX\252\002\rChalk.Auth.V1\312\002\rChalk\\Auth\\V1\342\002\031Chalk\\Auth\\V1\\GPBMetadata\352\002\017Chalk::Auth::V1"
    _globals["_FEATUREPERMISSIONS_TAGSENTRY"]._options = None
    _globals["_FEATUREPERMISSIONS_TAGSENTRY"]._serialized_options = b"8\001"
    _globals["_FEATUREPERMISSION"]._serialized_start = 236
    _globals["_FEATUREPERMISSION"]._serialized_end = 389
    _globals["_FEATUREPERMISSIONS"]._serialized_start = 57
    _globals["_FEATUREPERMISSIONS"]._serialized_end = 233
    _globals["_FEATUREPERMISSIONS_TAGSENTRY"]._serialized_start = 144
    _globals["_FEATUREPERMISSIONS_TAGSENTRY"]._serialized_end = 233
# @@protoc_insertion_point(module_scope)



from chalk._gen.chalk.artifacts.v1 import chart_pb2 as _chart_pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.chart.v1 import densetimeserieschart_pb2 as _densetimeserieschart_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class Series(_message.Message):
    __slots__ = ("points", "label", "units")
    POINTS_FIELD_NUMBER: _ClassVar[int]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    UNITS_FIELD_NUMBER: _ClassVar[int]
    points: _containers.RepeatedScalarFieldContainer[float]
    label: str
    units: str
    def __init__(
        self, points: _Optional[_Iterable[float]] = ..., label: _Optional[str] = ..., units: _Optional[str] = ...
    ) -> None: ...

class Chart(_message.Message):
    __slots__ = ("title", "series", "x_timestamp_ms")
    TITLE_FIELD_NUMBER: _ClassVar[int]
    SERIES_FIELD_NUMBER: _ClassVar[int]
    X_TIMESTAMP_MS_FIELD_NUMBER: _ClassVar[int]
    title: str
    series: _containers.RepeatedCompositeFieldContainer[Series]
    x_timestamp_ms: _containers.RepeatedScalarFieldContainer[int]
    def __init__(
        self,
        title: _Optional[str] = ...,
        series: _Optional[_Iterable[_Union[Series, _Mapping]]] = ...,
        x_timestamp_ms: _Optional[_Iterable[int]] = ...,
    ) -> None: ...

class Point(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: int
    def __init__(self, value: _Optional[int] = ...) -> None: ...

class TimeSeries(_message.Message):
    __slots__ = ("points", "label", "units")
    POINTS_FIELD_NUMBER: _ClassVar[int]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    UNITS_FIELD_NUMBER: _ClassVar[int]
    points: _containers.RepeatedCompositeFieldContainer[Point]
    label: str
    units: str
    def __init__(
        self,
        points: _Optional[_Iterable[_Union[Point, _Mapping]]] = ...,
        label: _Optional[str] = ...,
        units: _Optional[str] = ...,
    ) -> None: ...

class TimeSeriesChart(_message.Message):
    __slots__ = ("title", "series", "x_series", "window_period")
    TITLE_FIELD_NUMBER: _ClassVar[int]
    SERIES_FIELD_NUMBER: _ClassVar[int]
    X_SERIES_FIELD_NUMBER: _ClassVar[int]
    WINDOW_PERIOD_FIELD_NUMBER: _ClassVar[int]
    title: str
    series: _containers.RepeatedCompositeFieldContainer[TimeSeries]
    x_series: _containers.RepeatedCompositeFieldContainer[_timestamp_pb2.Timestamp]
    window_period: _duration_pb2.Duration
    def __init__(
        self,
        title: _Optional[str] = ...,
        series: _Optional[_Iterable[_Union[TimeSeries, _Mapping]]] = ...,
        x_series: _Optional[_Iterable[_Union[_timestamp_pb2.Timestamp, _Mapping]]] = ...,
        window_period: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
    ) -> None: ...

class ListChartsFilters(_message.Message):
    __slots__ = ("link_entity_kind", "linked_entity_id")
    LINK_ENTITY_KIND_FIELD_NUMBER: _ClassVar[int]
    LINKED_ENTITY_ID_FIELD_NUMBER: _ClassVar[int]
    link_entity_kind: _chart_pb2.ChartLinkKind
    linked_entity_id: str
    def __init__(
        self,
        link_entity_kind: _Optional[_Union[_chart_pb2.ChartLinkKind, str]] = ...,
        linked_entity_id: _Optional[str] = ...,
    ) -> None: ...

class ListChartPageToken(_message.Message):
    __slots__ = ("created_at_hwm", "id_hwm")
    CREATED_AT_HWM_FIELD_NUMBER: _ClassVar[int]
    ID_HWM_FIELD_NUMBER: _ClassVar[int]
    created_at_hwm: _timestamp_pb2.Timestamp
    id_hwm: str
    def __init__(
        self, created_at_hwm: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., id_hwm: _Optional[str] = ...
    ) -> None: ...

class ListChartsRequest(_message.Message):
    __slots__ = ("filters", "limit", "page_token")
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    filters: ListChartsFilters
    limit: int
    page_token: str
    def __init__(
        self,
        filters: _Optional[_Union[ListChartsFilters, _Mapping]] = ...,
        limit: _Optional[int] = ...,
        page_token: _Optional[str] = ...,
    ) -> None: ...

class ListChartsResponse(_message.Message):
    __slots__ = ("charts", "charts_with_links", "next_page_token")
    CHARTS_FIELD_NUMBER: _ClassVar[int]
    CHARTS_WITH_LINKS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    charts: _containers.RepeatedCompositeFieldContainer[_chart_pb2.MetricConfig]
    charts_with_links: _containers.RepeatedCompositeFieldContainer[_chart_pb2.Chart]
    next_page_token: str
    def __init__(
        self,
        charts: _Optional[_Iterable[_Union[_chart_pb2.MetricConfig, _Mapping]]] = ...,
        charts_with_links: _Optional[_Iterable[_Union[_chart_pb2.Chart, _Mapping]]] = ...,
        next_page_token: _Optional[str] = ...,
    ) -> None: ...

class GetChartSnapshotRequest(_message.Message):
    __slots__ = ("metric_config", "start_time", "end_time")
    METRIC_CONFIG_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    metric_config: _chart_pb2.MetricConfig
    start_time: _timestamp_pb2.Timestamp
    end_time: _timestamp_pb2.Timestamp
    def __init__(
        self,
        metric_config: _Optional[_Union[_chart_pb2.MetricConfig, _Mapping]] = ...,
        start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class GetChartSnapshotResponse(_message.Message):
    __slots__ = ("charts", "x_series", "window_period")
    CHARTS_FIELD_NUMBER: _ClassVar[int]
    X_SERIES_FIELD_NUMBER: _ClassVar[int]
    WINDOW_PERIOD_FIELD_NUMBER: _ClassVar[int]
    charts: _containers.RepeatedCompositeFieldContainer[_densetimeserieschart_pb2.DenseTimeSeriesChart]
    x_series: _containers.RepeatedCompositeFieldContainer[_timestamp_pb2.Timestamp]
    window_period: _duration_pb2.Duration
    def __init__(
        self,
        charts: _Optional[_Iterable[_Union[_densetimeserieschart_pb2.DenseTimeSeriesChart, _Mapping]]] = ...,
        x_series: _Optional[_Iterable[_Union[_timestamp_pb2.Timestamp, _Mapping]]] = ...,
        window_period: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
    ) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/offline_queries.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.aggregate.v1 import service_pb2 as chalk_dot_aggregate_dot_v1_dot_service__pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.server.v1 import datasets_pb2 as chalk_dot_server_dot_v1_dot_datasets__pb2
from chalk._gen.chalk.server.v1 import performance_summary_pb2 as chalk_dot_server_dot_v1_dot_performance__summary__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n%chalk/server/v1/offline_queries.proto\x12\x0f\x63halk.server.v1\x1a chalk/aggregate/v1/service.proto\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a\x1e\x63halk/server/v1/datasets.proto\x1a)chalk/server/v1/performance_summary.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\x8d\x12\n\x10OfflineQueryMeta\x12\x0e\n\x02id\x18\x01 \x01(\x03R\x02id\x12!\n\x0coperation_id\x18\x02 \x01(\tR\x0boperationId\x12%\n\x0e\x65nvironment_id\x18\x03 \x01(\tR\renvironmentId\x12#\n\rdeployment_id\x18\x04 \x01(\tR\x0c\x64\x65ploymentId\x12\x39\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12:\n\nquery_meta\x18\x06 \x01(\x0b\x32\x16.google.protobuf.ValueH\x00R\tqueryMeta\x88\x01\x01\x12\'\n\rquery_plan_id\x18\x07 \x01(\tH\x01R\x0bqueryPlanId\x88\x01\x01\x12$\n\x0b\x62ranch_name\x18\x08 \x01(\tH\x02R\nbranchName\x88\x01\x01\x12"\n\ndataset_id\x18\t \x01(\tH\x03R\tdatasetId\x88\x01\x01\x12&\n\x0c\x64\x61taset_name\x18\n \x01(\tH\x04R\x0b\x64\x61tasetName\x88\x01\x01\x12"\n\nhas_errors\x18\x0b \x01(\x08H\x05R\thasErrors\x88\x01\x01\x12\x1e\n\x08\x61gent_id\x18\x0c \x01(\tH\x06R\x07\x61gentId\x88\x01\x01\x12\x1e\n\x08trace_id\x18\r \x01(\tH\x07R\x07traceId\x88\x01\x01\x12*\n\x0e\x63orrelation_id\x18\x0e \x01(\tH\x08R\rcorrelationId\x88\x01\x01\x12=\n\x0c\x63ompleted_at\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0b\x63ompletedAt\x12@\n\x06status\x18\x10 \x01(\x0e\x32#.chalk.server.v1.OfflineQueryStatusH\tR\x06status\x88\x01\x01\x12&\n\x0fhas_plan_stages\x18\x11 \x01(\x08R\rhasPlanStages\x12,\n\x0ftotal_computers\x18\x12 \x01(\x05H\nR\x0etotalComputers\x88\x01\x01\x12;\n\x17num_completed_computers\x18\x13 \x01(\x05H\x0bR\x15numCompletedComputers\x88\x01\x01\x12.\n\x10total_partitions\x18\x14 \x01(\x05H\x0cR\x0ftotalPartitions\x88\x01\x01\x12=\n\x18num_completed_partitions\x18\x15 \x01(\x05H\rR\x16numCompletedPartitions\x88\x01\x01\x12\x32\n\x12recompute_features\x18\x16 \x01(\tH\x0eR\x11recomputeFeatures\x88\x01\x01\x12+\n\x0fspine_sql_query\x18\x17 \x01(\tH\x0fR\rspineSqlQuery\x88\x01\x01\x12\x35\n\x07\x66ilters\x18\x18 \x01(\x0b\x32\x16.google.protobuf.ValueH\x10R\x07\x66ilters\x88\x01\x01\x12\x44\n\x0fplanner_options\x18\x19 \x01(\x0b\x32\x16.google.protobuf.ValueH\x11R\x0eplannerOptions\x88\x01\x01\x12\x44\n\x0finvoker_options\x18\x1a \x01(\x0b\x32\x16.google.protobuf.ValueH\x12R\x0einvokerOptions\x88\x01\x01\x12\x45\n\nquery_type\x18\x1b \x01(\x0e\x32!.chalk.server.v1.OfflineQueryKindH\x13R\tqueryType\x88\x01\x01\x12\x12\n\x04tags\x18\x1c \x03(\tR\x04tags\x12\x34\n\x16required_resolver_tags\x18\x1d \x03(\tR\x14requiredResolverTags\x12\x37\n\x15\x61ggregate_backfill_id\x18\x1e \x01(\tH\x14R\x13\x61ggregateBackfillId\x88\x01\x01\x12\x33\n\x06output\x18\x1f \x01(\x0b\x32\x16.google.protobuf.ValueH\x15R\x06output\x88\x01\x01\x12\x44\n\x0frequired_output\x18  \x01(\x0b\x32\x16.google.protobuf.ValueH\x16R\x0erequiredOutput\x88\x01\x01\x12/\n\x11raw_body_filename\x18! \x01(\tH\x17R\x0frawBodyFilename\x88\x01\x01\x12T\n\x10\x64\x61taset_revision\x18" \x01(\x0b\x32$.chalk.server.v1.DatasetRevisionMetaH\x18R\x0f\x64\x61tasetRevision\x88\x01\x01\x12R\n\x0btime_series\x18# \x03(\x0b\x32\x31.chalk.aggregate.v1.PlanAggregateBackfillResponseR\ntimeSeries\x12/\n\x11\x65valuation_run_id\x18$ \x01(\tH\x19R\x0f\x65valuationRunId\x88\x01\x01\x42\r\n\x0b_query_metaB\x10\n\x0e_query_plan_idB\x0e\n\x0c_branch_nameB\r\n\x0b_dataset_idB\x0f\n\r_dataset_nameB\r\n\x0b_has_errorsB\x0b\n\t_agent_idB\x0b\n\t_trace_idB\x11\n\x0f_correlation_idB\t\n\x07_statusB\x12\n\x10_total_computersB\x1a\n\x18_num_completed_computersB\x13\n\x11_total_partitionsB\x1b\n\x19_num_completed_partitionsB\x15\n\x13_recompute_featuresB\x12\n\x10_spine_sql_queryB\n\n\x08_filtersB\x12\n\x10_planner_optionsB\x12\n\x10_invoker_optionsB\r\n\x0b_query_typeB\x18\n\x16_aggregate_backfill_idB\t\n\x07_outputB\x12\n\x10_required_outputB\x14\n\x12_raw_body_filenameB\x13\n\x11_dataset_revisionB\x14\n\x12_evaluation_run_id"\xf7\x05\n\x19ListOfflineQueriesRequest\x12\x1b\n\x06\x63ursor\x18\x01 \x01(\tH\x00R\x06\x63ursor\x88\x01\x01\x12\x19\n\x05limit\x18\x02 \x01(\x05H\x01R\x05limit\x88\x01\x01\x12"\n\nstart_date\x18\x03 \x01(\tH\x02R\tstartDate\x88\x01\x01\x12\x1e\n\x08\x65nd_date\x18\x04 \x01(\tH\x03R\x07\x65ndDate\x88\x01\x01\x12 \n\tid_filter\x18\x05 \x01(\tH\x04R\x08idFilter\x88\x01\x01\x12+\n\x0f\x61gent_id_filter\x18\x06 \x01(\tH\x05R\ragentIdFilter\x88\x01\x01\x12(\n\rbranch_filter\x18\x07 \x01(\tH\x06R\x0c\x62ranchFilter\x88\x01\x01\x12G\n\x0bkind_filter\x18\x08 \x01(\x0e\x32!.chalk.server.v1.OfflineQueryKindH\x07R\nkindFilter\x88\x01\x01\x12M\n\rstatus_filter\x18\t \x01(\x0e\x32#.chalk.server.v1.OfflineQueryStatusH\x08R\x0cstatusFilter\x88\x01\x01\x12H\n\x1e\x61ggregation_backfill_id_filter\x18\n \x01(\tH\tR\x1b\x61ggregationBackfillIdFilter\x88\x01\x01\x12<\n\x18\x65valuation_run_id_filter\x18\x0b \x01(\tH\nR\x15\x65valuationRunIdFilter\x88\x01\x01\x42\t\n\x07_cursorB\x08\n\x06_limitB\r\n\x0b_start_dateB\x0b\n\t_end_dateB\x0c\n\n_id_filterB\x12\n\x10_agent_id_filterB\x10\n\x0e_branch_filterB\x0e\n\x0c_kind_filterB\x10\n\x0e_status_filterB!\n\x1f_aggregation_backfill_id_filterB\x1b\n\x19_evaluation_run_id_filter"\x90\x01\n\x1aListOfflineQueriesResponse\x12J\n\x0foffline_queries\x18\x01 \x03(\x0b\x32!.chalk.server.v1.OfflineQueryMetaR\x0eofflineQueries\x12\x1b\n\x06\x63ursor\x18\x02 \x01(\tH\x00R\x06\x63ursor\x88\x01\x01\x42\t\n\x07_cursor"B\n\x16GetOfflineQueryRequest\x12(\n\x10offline_query_id\x18\x01 \x01(\tR\x0eofflineQueryId"a\n\x17GetOfflineQueryResponse\x12\x46\n\roffline_query\x18\x01 \x01(\x0b\x32!.chalk.server.v1.OfflineQueryMetaR\x0cofflineQuery*\x87\x02\n\x12OfflineQueryStatus\x12$\n OFFLINE_QUERY_STATUS_UNSPECIFIED\x10\x00\x12 \n\x1cOFFLINE_QUERY_STATUS_UNKNOWN\x10\x01\x12 \n\x1cOFFLINE_QUERY_STATUS_WORKING\x10\x02\x12\x1f\n\x1bOFFLINE_QUERY_STATUS_FAILED\x10\x03\x12"\n\x1eOFFLINE_QUERY_STATUS_COMPLETED\x10\x04\x12!\n\x1dOFFLINE_QUERY_STATUS_CANCELED\x10\x05\x12\x1f\n\x1bOFFLINE_QUERY_STATUS_QUEUED\x10\x06*\xaa\x02\n\x10OfflineQueryKind\x12"\n\x1eOFFLINE_QUERY_KIND_UNSPECIFIED\x10\x00\x12\x1e\n\x1aOFFLINE_QUERY_KIND_UNKNOWN\x10\x01\x12*\n&OFFLINE_QUERY_KIND_ASYNC_OFFLINE_QUERY\x10\x02\x12)\n%OFFLINE_QUERY_KIND_CRON_OFFLINE_QUERY\x10\x03\x12$\n OFFLINE_QUERY_KIND_OFFLINE_QUERY\x10\x04\x12(\n$OFFLINE_QUERY_KIND_DATASET_INGESTION\x10\x05\x12+\n\'OFFLINE_QUERY_KIND_AGGREGATION_BACKFILL\x10\x06\x32\xbf\x03\n\x1bOfflineQueryMetadataService\x12u\n\x12ListOfflineQueries\x12*.chalk.server.v1.ListOfflineQueriesRequest\x1a+.chalk.server.v1.ListOfflineQueriesResponse"\x06\x90\x02\x01\x80}\x0b\x12l\n\x0fGetOfflineQuery\x12\'.chalk.server.v1.GetOfflineQueryRequest\x1a(.chalk.server.v1.GetOfflineQueryResponse"\x06\x90\x02\x01\x80}\x0b\x12\xba\x01\n)ListOfflineQueryShardPerformanceSummaries\x12\x41.chalk.server.v1.ListOfflineQueryShardPerformanceSummariesRequest\x1a\x42.chalk.server.v1.ListOfflineQueryShardPerformanceSummariesResponse"\x06\x90\x02\x01\x80}\x0b\x42\x9c\x01\n\x13\x63om.chalk.server.v1B\x13OfflineQueriesProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.offline_queries_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\023OfflineQueriesProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_OFFLINEQUERYMETADATASERVICE"].methods_by_name["ListOfflineQueries"]._options = None
    _globals["_OFFLINEQUERYMETADATASERVICE"].methods_by_name[
        "ListOfflineQueries"
    ]._serialized_options = b"\220\002\001\200}\013"
    _globals["_OFFLINEQUERYMETADATASERVICE"].methods_by_name["GetOfflineQuery"]._options = None
    _globals["_OFFLINEQUERYMETADATASERVICE"].methods_by_name[
        "GetOfflineQuery"
    ]._serialized_options = b"\220\002\001\200}\013"
    _globals["_OFFLINEQUERYMETADATASERVICE"].methods_by_name[
        "ListOfflineQueryShardPerformanceSummaries"
    ]._options = None
    _globals["_OFFLINEQUERYMETADATASERVICE"].methods_by_name[
        "ListOfflineQueryShardPerformanceSummaries"
    ]._serialized_options = b"\220\002\001\200}\013"
    _globals["_OFFLINEQUERYSTATUS"]._serialized_start = 3660
    _globals["_OFFLINEQUERYSTATUS"]._serialized_end = 3923
    _globals["_OFFLINEQUERYKIND"]._serialized_start = 3926
    _globals["_OFFLINEQUERYKIND"]._serialized_end = 4224
    _globals["_OFFLINEQUERYMETA"]._serialized_start = 264
    _globals["_OFFLINEQUERYMETA"]._serialized_end = 2581
    _globals["_LISTOFFLINEQUERIESREQUEST"]._serialized_start = 2584
    _globals["_LISTOFFLINEQUERIESREQUEST"]._serialized_end = 3343
    _globals["_LISTOFFLINEQUERIESRESPONSE"]._serialized_start = 3346
    _globals["_LISTOFFLINEQUERIESRESPONSE"]._serialized_end = 3490
    _globals["_GETOFFLINEQUERYREQUEST"]._serialized_start = 3492
    _globals["_GETOFFLINEQUERYREQUEST"]._serialized_end = 3558
    _globals["_GETOFFLINEQUERYRESPONSE"]._serialized_start = 3560
    _globals["_GETOFFLINEQUERYRESPONSE"]._serialized_end = 3657
    _globals["_OFFLINEQUERYMETADATASERVICE"]._serialized_start = 4227
    _globals["_OFFLINEQUERYMETADATASERVICE"]._serialized_end = 4674
# @@protoc_insertion_point(module_scope)



from typing import Any, Mapping, Union

from google.protobuf.message import Message

from chalk.features._encoding.protobuf import (
    create_empty_pyarrow_scalar_from_proto_type,
    serialize_message_file_descriptor,
)
from chalk.features.underscore import Underscore, UnderscoreFunction


def _is_protobuf_message(obj: Any) -> bool:
    if isinstance(obj, Message):
        return True

    # If using a different protobuf generation implemention, e.g. google._upb._message.MessageMeta,
    # just check for common protobuf fields fields
    required_methods = [
        "DESCRIPTOR",
        "SerializeToString",
        "ParseFromString",
    ]
    return all(hasattr(obj, method) for method in required_methods)


class UnderscoreProtoSerialize(UnderscoreFunction):
    def __init__(
        self,
        mapping: Mapping[str, Union[Underscore, Any]],
        message: Message,
    ):
        if not isinstance(mapping, Mapping):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"F.proto_serialize(): mapping must be a Mapping: got {type(mapping)}")
        if not _is_protobuf_message(message):
            raise TypeError(f"F.proto_serialize(): message must be a Message: got {type(message)}")

        message_file_descriptor = serialize_message_file_descriptor(message.DESCRIPTOR.file)
        super().__init__(
            "proto_serialize",
            message_file_descriptor,
            message.DESCRIPTOR.full_name,
            list(mapping.keys()),
            *mapping.values(),
        )


def proto_serialize(mapping: Mapping[str, Union[Underscore, Any]], message: Message):
    """
    Serialize a proto message from a mapping of field names to values.

    Parameters
    ----------
    mapping
        The mapping of names to features to serialize.
    message
        The proto message to serialize.

    Examples
    --------
    >>> import chalk.functions as F
    >>> from chalk.features import _, features
    >>> from protos.gen.v1.transaction_pb2 import GetTransactionRequest
    >>> @features
    ... class Transaction:
    ...    id: int
    ...    transaction_request: bytes = F.proto_serialize(
    ...        {
    ...            "id": _.id,
    ...        },
    ...        GetTransactionRequest,
    ...    )
    """
    return UnderscoreProtoSerialize(mapping, message)


class UnderscoreProtoDeserialize(UnderscoreFunction):
    def __init__(
        self,
        body: Union[Underscore, bytes],
        message: Message,
    ):
        if not isinstance(body, (bytes, Underscore)):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"F.proto_deserialize(): body must be a bytes or Underscore, got {type(body)}")
        if not _is_protobuf_message(message):
            raise TypeError(f"F.proto_deserialize(): message must be a Message, got {type(message)}")

        message_file_descriptor = serialize_message_file_descriptor(message.DESCRIPTOR.file)
        message_name = message.DESCRIPTOR.full_name
        pa_scalar = create_empty_pyarrow_scalar_from_proto_type(message)
        super().__init__(
            "proto_deserialize",
            message_file_descriptor,
            message_name,
            pa_scalar,
            body,
        )


def proto_deserialize(body: Union[Underscore, bytes], message: Message):
    """
    Deserialize a proto message from a bytes feature.

    Parameters
    ----------
    body
        The bytes feature to deserialize.
    message
        The proto message type to deserialize.

    Examples
    --------
    >>> import chalk.functions as F
    >>> from chalk.features import _, features
    >>> from protos.gen.v1.transaction_pb2 import GetTransactionResponse
    >>> @features
    ... class Transaction:
    ...    id: int
    ...    transaction_response_bytes: bytes
    ...    transaction_response: GetTransactionResponse = F.proto_deserialize(
    ...        _.transaction_response_bytes,
    ...        GetTransactionResponse,
    ...    )
    """
    return UnderscoreProtoDeserialize(body, message)


__all__ = [
    "proto_serialize",
    "proto_deserialize",
]



import base64
import collections.abc
import traceback
from datetime import datetime, timedelta
from typing import Any, Optional, Type, TypeVar, Union, cast

import google.protobuf.duration_pb2 as duration_pb
from google.protobuf import duration_pb2, message, struct_pb2, timestamp_pb2

import chalk._gen.chalk.artifacts.v1.export_pb2 as export_pb
import chalk._gen.chalk.graph.v1.graph_pb2 as graph_pb
import chalk._gen.chalk.graph.v2.sources_pb2 as sources_pb
from chalk.parsed.duplicate_input_gql import FailedImport
from chalk.utils.duration import CHALK_MAX_TIMEDELTA
from chalk.utils.json import TJSON

RESOLVER_ENUM_TO_KIND = {
    graph_pb.RESOLVER_KIND_UNSPECIFIED: "UNSPECIFIED",
    graph_pb.RESOLVER_KIND_ONLINE: "online",
    graph_pb.RESOLVER_KIND_OFFLINE: "offline",
}


def encode_proto_to_b64(obj: message.Message, deterministic: bool) -> str:
    b = obj.SerializeToString(deterministic=deterministic)
    return base64.b64encode(b).decode("utf-8")


T = TypeVar("T", bound=message.Message)


def decode_proto_from_b64(b64: str, proto_cls: Type[T]) -> T:
    proto_obj = proto_cls()
    proto_obj.ParseFromString(base64.b64decode(b64))
    return proto_obj


def build_failed_import(error: Union[Exception, str], description: str) -> export_pb.FailedImport:
    try:
        formatted_tb = error if isinstance(error, str) else "\n".join(traceback.format_exception(error))
    except:
        formatted_tb = (
            error
            if isinstance(error, str)
            else "\n".join(
                traceback.format_exception(
                    type(error),
                    value=error,
                    tb=None,
                )
            )
        )

    return export_pb.FailedImport(
        file_name="",
        module="",
        traceback=f"EXCEPTION in parsing {description}:\n{formatted_tb}",
    )


def convert_failed_import_to_proto(failed_import: FailedImport) -> export_pb.FailedImport:
    return export_pb.FailedImport(
        file_name=failed_import.filename,
        module=failed_import.module,
        traceback=failed_import.traceback,
    )


def convert_failed_import_to_gql(failed_import: export_pb.FailedImport) -> FailedImport:
    return FailedImport(
        filename=failed_import.file_name,
        module=failed_import.module,
        traceback=failed_import.traceback,
    )


def get_feature_type_attribute_name(feature: graph_pb.FeatureType) -> str:
    if feature.HasField("group_by"):
        return feature.group_by.attribute_name
    elif feature.HasField("scalar"):
        return feature.scalar.attribute_name
    elif feature.HasField("has_one"):
        return feature.has_one.attribute_name
    elif feature.HasField("has_many"):
        return feature.has_many.attribute_name
    elif feature.HasField("feature_time"):
        return feature.feature_time.attribute_name
    elif feature.HasField("windowed"):
        return feature.windowed.attribute_name
    else:
        raise ValueError(f"Unknown feature type: {feature}")


def get_feature_type_fqn(feature: graph_pb.FeatureType) -> str:
    if feature.HasField("group_by"):
        return f"{feature.group_by.namespace}.{feature.group_by.name}"
    if feature.HasField("scalar"):
        return f"{feature.scalar.namespace}.{feature.scalar.name}"
    elif feature.HasField("has_one"):
        return f"{feature.has_one.namespace}.{feature.has_one.name}"
    elif feature.HasField("has_many"):
        return f"{feature.has_many.namespace}.{feature.has_many.name}"
    elif feature.HasField("feature_time"):
        return f"{feature.feature_time.namespace}.{feature.feature_time.name}"
    elif feature.HasField("windowed"):
        return f"{feature.windowed.namespace}.{feature.windowed.name}"
    else:
        raise ValueError(f"Unknown feature type: {feature}")


def get_database_source_name(source: sources_pb.DatabaseSource) -> str:
    return source.name


def get_stream_source_name(source: sources_pb.StreamSource) -> str:
    return source.name


def get_feature_reference_fqn(feature: graph_pb.FeatureReference) -> str:
    return f"{feature.namespace}.{feature.name}"


def get_feature_type_name(feature: graph_pb.FeatureType) -> str:
    if feature.HasField("group_by"):
        return feature.group_by.name
    elif feature.HasField("scalar"):
        return feature.scalar.name
    elif feature.HasField("has_one"):
        return feature.has_one.name
    elif feature.HasField("has_many"):
        return feature.has_many.name
    elif feature.HasField("feature_time"):
        return feature.feature_time.name
    elif feature.HasField("windowed"):
        return feature.windowed.name
    else:
        raise ValueError(f"Unknown feature type: {feature}")


def get_feature_type_namespace(feature: graph_pb.FeatureType) -> str:
    if feature.HasField("group_by"):
        return feature.group_by.namespace
    elif feature.HasField("scalar"):
        return feature.scalar.namespace
    elif feature.HasField("has_one"):
        return feature.has_one.namespace
    elif feature.HasField("has_many"):
        return feature.has_many.namespace
    elif feature.HasField("feature_time"):
        return feature.feature_time.namespace
    elif feature.HasField("windowed"):
        return feature.windowed.namespace
    else:
        raise ValueError(f"Unknown feature type: {feature}")


def get_feature_type_etl_offline_to_online(feature: graph_pb.FeatureType) -> Optional[bool]:
    if feature.HasField("group_by"):
        return None
    elif feature.HasField("scalar"):
        return feature.scalar.etl_offline_to_online if feature.scalar.HasField("etl_offline_to_online") else None
    elif feature.HasField("has_one"):
        return None
    elif feature.HasField("has_many"):
        return None
    elif feature.HasField("feature_time"):
        return None
    elif feature.HasField("windowed"):
        return None
    else:
        raise ValueError(f"Unknown feature type: {feature}")


def get_feature_type_max_staleness(feature: graph_pb.FeatureType) -> Optional[duration_pb.Duration]:
    if feature.HasField("group_by"):
        return None
    elif feature.HasField("scalar"):
        return feature.scalar.max_staleness_duration if feature.scalar.HasField("max_staleness_duration") else None
    elif feature.HasField("has_one"):
        return None
    elif feature.HasField("has_many"):
        return feature.has_many.max_staleness_duration if feature.has_many.HasField("max_staleness_duration") else None
    elif feature.HasField("feature_time"):
        return None
    elif feature.HasField("windowed"):
        return None
    else:
        raise ValueError(f"Unknown feature type: {feature}")


def get_feature_type_is_autogenerated(feature: graph_pb.FeatureType) -> bool:
    if feature.HasField("group_by"):
        return False
    elif feature.HasField("scalar"):
        return feature.scalar.is_autogenerated
    elif feature.HasField("has_one"):
        return feature.has_one.is_autogenerated
    elif feature.HasField("has_many"):
        return feature.has_many.is_autogenerated
    elif feature.HasField("feature_time"):
        return feature.feature_time.is_autogenerated
    elif feature.HasField("windowed"):
        return feature.windowed.is_autogenerated
    else:
        raise ValueError(f"Unknown feature type: {feature}")


_max_timedelta_seconds = CHALK_MAX_TIMEDELTA.total_seconds()


def seconds_to_proto_duration(seconds: float) -> duration_pb2.Duration:
    return timedelta_to_proto_duration(
        CHALK_MAX_TIMEDELTA if seconds >= _max_timedelta_seconds else timedelta(seconds=seconds)
    )


def seconds_int_to_proto_duration(seconds: int) -> duration_pb2.Duration:
    seconds = min(CHALK_MAX_TIMEDELTA.seconds + CHALK_MAX_TIMEDELTA.days * 86400, seconds)
    pb_duration = duration_pb2.Duration()
    pb_duration.FromSeconds(seconds)
    return pb_duration


def timedelta_to_proto_duration(duration: timedelta) -> duration_pb2.Duration:
    pb_duration = duration_pb2.Duration()
    if duration >= CHALK_MAX_TIMEDELTA:
        pb_duration.seconds = CHALK_MAX_TIMEDELTA.seconds + CHALK_MAX_TIMEDELTA.days * 86400
        pb_duration.nanos = CHALK_MAX_TIMEDELTA.microseconds * 1000
        return pb_duration
    try:
        pb_duration.FromTimedelta(duration)
        return pb_duration
    except Exception as e:
        raise ValueError(f"Invalid duration: {e}")


def proto_duration_to_timedelta(duration_proto: duration_pb2.Duration) -> timedelta:
    return duration_proto.ToTimedelta()


def datetime_to_proto_timestamp(dt: datetime) -> timestamp_pb2.Timestamp:
    pb_timestamp = timestamp_pb2.Timestamp()
    pb_timestamp.FromDatetime(dt)
    return pb_timestamp


def value_to_proto(py_obj: Optional[TJSON]) -> struct_pb2.Value:
    if py_obj is None:
        return struct_pb2.Value(null_value=struct_pb2.NullValue.NULL_VALUE)
    elif isinstance(py_obj, str):
        return struct_pb2.Value(string_value=py_obj)
    elif isinstance(py_obj, bool):
        # IMPORTANT: This check needs to be before the int check, since isinstance(True, int) evaluates to True :-(
        return struct_pb2.Value(bool_value=py_obj)
    elif isinstance(py_obj, (int, float)):
        return struct_pb2.Value(number_value=py_obj)
    elif isinstance(py_obj, collections.abc.Mapping):
        s = struct_pb2.Struct()
        s.update(py_obj)
        return struct_pb2.Value(struct_value=s)
    elif isinstance(py_obj, collections.abc.Sequence):  # pyright: ignore[reportUnnecessaryIsInstance]
        l = struct_pb2.ListValue()
        l.extend(cast(Any, py_obj))
        return struct_pb2.Value(list_value=l)
    else:
        raise TypeError(f"Unsupported python type for conversion to protobuf struct_pb2.Value: {type(py_obj)}")


def proto_value_to_python(value_proto: Optional[struct_pb2.Value]) -> Optional[TJSON]:
    if value_proto is None:
        return None
    kind = value_proto.WhichOneof("kind")
    if kind is None:
        # This is needed for empty/un-set Value's, although technically the caller should have set it with a 'null_value'.
        return None
    if kind == "null_value":
        return None
    elif kind == "number_value":
        return value_proto.number_value
    elif kind == "string_value":
        return value_proto.string_value
    elif kind == "bool_value":
        return value_proto.bool_value
    elif kind == "struct_value":
        return {k: proto_value_to_python(v) for k, v in value_proto.struct_value.fields.items()}
    elif kind == "list_value":
        return [proto_value_to_python(v) for v in value_proto.list_value.values]
    else:
        raise TypeError(f"Unsupported struct_pb2.kind value: '{kind}'")



import functools

from chalk.utils.environment_parsing import env_var_bool


@functools.lru_cache(None)
def should_skip_source_code_parsing():
    """Whether to skip parsing source code for better error messages. Setting `CHALK_DISABLE_SOURCE_CODE_PARSING` to True can help improve graph loading performance."""
    return env_var_bool("CHALK_DISABLE_SOURCE_CODE_PARSING")



import dataclasses
from typing import TYPE_CHECKING, Any, Mapping, Union

from rich.text import Text

from chalk.client import ErrorCode
from chalk.client.models import ChalkError
from chalk.features.feature_field import Feature, FeatureWrapper
from chalk.utils.json import TJSON

if TYPE_CHECKING:
    import polars as pl


@dataclasses.dataclass
class ManyQueryInputsCheckerWrapper:
    inputs: Mapping[Union[str, FeatureWrapper, Feature, Any], list[Any]]


@dataclasses.dataclass(frozen=True)
class Result:
    fqn: str
    value: Union[TJSON, "pl.DataFrame"]
    cache_hit: bool
    error: Union[ChalkError, ErrorCode, None]
    pkey: TJSON = None


class Color:
    R = "\033[031m"  # RED
    G = "\033[032m"  # GREEN
    Y = "\033[033m"  # Yellow
    B = "\033[034m"  # Blue
    N = "\033[0m"  # Reset

    @staticmethod
    def render(s: str, color: str):
        return Text.from_ansi(f"{color}{s}{Color.N}")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/engine/v2/feature_values.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.arrow.v1 import arrow_pb2 as chalk_dot_arrow_dot_v1_dot_arrow__pb2
from chalk._gen.chalk.common.v1 import operation_kind_pb2 as chalk_dot_common_dot_v1_dot_operation__kind__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n$chalk/engine/v2/feature_values.proto\x12\x0f\x63halk.engine.v2\x1a\x1a\x63halk/arrow/v1/arrow.proto\x1a$chalk/common/v1/operation_kind.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\x85\x02\n\x13\x46\x65\x61tureValueFilters\x12!\n\x0cresolver_fqn\x18\x01 \x03(\tR\x0bresolverFqn\x12#\n\rdeployment_id\x18\x02 \x03(\tR\x0c\x64\x65ploymentId\x12!\n\x0coperation_id\x18\x03 \x03(\tR\x0boperationId\x12\x45\n\x0eoperation_kind\x18\x04 \x03(\x0e\x32\x1e.chalk.common.v1.OperationKindR\roperationKind\x12<\n\x0bprimary_key\x18\x05 \x03(\x0b\x32\x1b.chalk.arrow.v1.ScalarValueR\nprimaryKey"\xb4\x01\n\x19GetFeatureValuesPageToken\x12?\n\rtimestamp_hwm\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0ctimestampHwm\x12(\n\x10operation_id_hwm\x18\x02 \x01(\tR\x0eoperationIdHwm\x12,\n\x12observation_id_hwm\x18\x03 \x01(\tR\x10observationIdHwm"\xaa\x03\n\x17GetFeatureValuesRequest\x12\x1f\n\x0b\x66\x65\x61ture_fqn\x18\x01 \x01(\tR\nfeatureFqn\x12R\n\x0etimestamp_type\x18\x02 \x01(\x0e\x32+.chalk.engine.v2.FeatureValuesTimestampTypeR\rtimestampType\x12N\n\x15lower_bound_inclusive\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x13lowerBoundInclusive\x12N\n\x15upper_bound_exclusive\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x13upperBoundExclusive\x12>\n\x07\x66ilters\x18\x05 \x01(\x0b\x32$.chalk.engine.v2.FeatureValueFiltersR\x07\x66ilters\x12\x1b\n\tpage_size\x18\x06 \x01(\x05R\x08pageSize\x12\x1d\n\npage_token\x18\x07 \x01(\tR\tpageToken"\x88\x01\n\x18GetFeatureValuesResponse\x12&\n\x0fnext_page_token\x18\x01 \x01(\tR\rnextPageToken\x12\x1d\n\ntotal_size\x18\x02 \x01(\x05R\ttotalSize\x12\x1a\n\x07parquet\x18\x03 \x01(\x0cH\x00R\x07parquetB\t\n\x07payload*\xa9\x01\n\x1a\x46\x65\x61tureValuesTimestampType\x12-\n)FEATURE_VALUES_TIMESTAMP_TYPE_UNSPECIFIED\x10\x00\x12-\n)FEATURE_VALUES_TIMESTAMP_TYPE_INSERTED_AT\x10\x01\x12-\n)FEATURE_VALUES_TIMESTAMP_TYPE_OBSERVED_AT\x10\x02\x42\x87\x01\n\x13\x63om.chalk.engine.v2B\x12\x46\x65\x61tureValuesProtoP\x01\xa2\x02\x03\x43\x45X\xaa\x02\x0f\x43halk.Engine.V2\xca\x02\x0f\x43halk\\Engine\\V2\xe2\x02\x1b\x43halk\\Engine\\V2\\GPBMetadata\xea\x02\x11\x43halk::Engine::V2b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.engine.v2.feature_values_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.engine.v2B\022FeatureValuesProtoP\001\242\002\003CEX\252\002\017Chalk.Engine.V2\312\002\017Chalk\\Engine\\V2\342\002\033Chalk\\Engine\\V2\\GPBMetadata\352\002\021Chalk::Engine::V2"
    _globals["_FEATUREVALUESTIMESTAMPTYPE"]._serialized_start = 1172
    _globals["_FEATUREVALUESTIMESTAMPTYPE"]._serialized_end = 1341
    _globals["_FEATUREVALUEFILTERS"]._serialized_start = 157
    _globals["_FEATUREVALUEFILTERS"]._serialized_end = 418
    _globals["_GETFEATUREVALUESPAGETOKEN"]._serialized_start = 421
    _globals["_GETFEATUREVALUESPAGETOKEN"]._serialized_end = 601
    _globals["_GETFEATUREVALUESREQUEST"]._serialized_start = 604
    _globals["_GETFEATUREVALUESREQUEST"]._serialized_end = 1030
    _globals["_GETFEATUREVALUESRESPONSE"]._serialized_start = 1033
    _globals["_GETFEATUREVALUESRESPONSE"]._serialized_end = 1169
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.auth_pb2 import (
    GetTokenRequest,
    GetTokenResponse,
    UpdateLinkSessionRequest,
    UpdateLinkSessionResponse,
)
from chalk._gen.chalk.server.v1.link_pb2 import (
    CreateLinkSessionRequest,
    CreateLinkSessionResponse,
    GetLinkSessionRequest,
    GetLinkSessionResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class AuthServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetToken: UnaryUnaryMultiCallable[
        GetTokenRequest,
        GetTokenResponse,
    ]
    CreateLinkSession: UnaryUnaryMultiCallable[
        CreateLinkSessionRequest,
        CreateLinkSessionResponse,
    ]
    GetLinkSession: UnaryUnaryMultiCallable[
        GetLinkSessionRequest,
        GetLinkSessionResponse,
    ]
    UpdateLinkSession: UnaryUnaryMultiCallable[
        UpdateLinkSessionRequest,
        UpdateLinkSessionResponse,
    ]

class AuthServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetToken(
        self,
        request: GetTokenRequest,
        context: ServicerContext,
    ) -> GetTokenResponse: ...
    @abstractmethod
    def CreateLinkSession(
        self,
        request: CreateLinkSessionRequest,
        context: ServicerContext,
    ) -> CreateLinkSessionResponse: ...
    @abstractmethod
    def GetLinkSession(
        self,
        request: GetLinkSessionRequest,
        context: ServicerContext,
    ) -> GetLinkSessionResponse: ...
    @abstractmethod
    def UpdateLinkSession(
        self,
        request: UpdateLinkSessionRequest,
        context: ServicerContext,
    ) -> UpdateLinkSessionResponse: ...

def add_AuthServiceServicer_to_server(servicer: AuthServiceServicer, server: Server) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/kube.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.kubernetes.v1 import events_pb2 as chalk_dot_kubernetes_dot_v1_dot_events__pb2
from chalk._gen.chalk.kubernetes.v1 import persistentvolume_pb2 as chalk_dot_kubernetes_dot_v1_dot_persistentvolume__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1a\x63halk/server/v1/kube.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a chalk/kubernetes/v1/events.proto\x1a*chalk/kubernetes/v1/persistentvolume.proto"\xb2\x02\n\x1bGetPodStackTraceDumpRequest\x12!\n\tnamespace\x18\x01 \x01(\tH\x01R\tnamespace\x88\x01\x01\x12\x19\n\x08pod_name\x18\x02 \x01(\tR\x07podName\x12*\n\x0e\x63ontainer_name\x18\x03 \x01(\tH\x02R\rcontainerName\x88\x01\x01\x12\x1f\n\nprocess_id\x18\x04 \x01(\x05H\x00R\tprocessId\x12#\n\x0cprocess_name\x18\x05 \x01(\tH\x00R\x0bprocessName\x12\x30\n\x13\x61uto_detect_process\x18\x06 \x01(\x08H\x00R\x11\x61utoDetectProcessB\x10\n\x0eprocess_filterB\x0c\n\n_namespaceB\x11\n\x0f_container_name"?\n\x1cGetPodStackTraceDumpResponse\x12\x1f\n\x0bstack_trace\x18\x01 \x01(\tR\nstackTrace"\x84\x02\n\x1aGetKubernetesEventsRequest\x12&\n\x0c\x63luster_name\x18\x01 \x01(\tH\x00R\x0b\x63lusterName\x88\x01\x01\x12!\n\tnamespace\x18\x02 \x01(\tH\x01R\tnamespace\x88\x01\x01\x12*\n\x0elabel_selector\x18\x03 \x01(\tH\x02R\rlabelSelector\x88\x01\x01\x12*\n\x0e\x66ield_selector\x18\x04 \x01(\tH\x03R\rfieldSelector\x88\x01\x01\x42\x0f\n\r_cluster_nameB\x0c\n\n_namespaceB\x11\n\x0f_label_selectorB\x11\n\x0f_field_selector"`\n\x1bGetKubernetesEventsResponse\x12\x41\n\x06\x65vents\x18\x01 \x03(\x0b\x32).chalk.kubernetes.v1.ChalkKubernetesEventR\x06\x65vents"\'\n%GetKubernetesPersistentVolumesRequest"x\n&GetKubernetesPersistentVolumesResponse\x12N\n\x07volumes\x18\x01 \x03(\x0b\x32\x34.chalk.kubernetes.v1.ChalkKubernetesPersistentVolumeR\x07volumes2\xa0\x03\n\x0bKubeService\x12{\n\x14GetPodStackTraceDump\x12,.chalk.server.v1.GetPodStackTraceDumpRequest\x1a-.chalk.server.v1.GetPodStackTraceDumpResponse"\x06\x90\x02\x01\x80}\x0b\x12x\n\x13GetKubernetesEvents\x12+.chalk.server.v1.GetKubernetesEventsRequest\x1a,.chalk.server.v1.GetKubernetesEventsResponse"\x06\x90\x02\x01\x80}\x0b\x12\x99\x01\n\x1eGetKubernetesPersistentVolumes\x12\x36.chalk.server.v1.GetKubernetesPersistentVolumesRequest\x1a\x37.chalk.server.v1.GetKubernetesPersistentVolumesResponse"\x06\x90\x02\x01\x80}\x0b\x42\x92\x01\n\x13\x63om.chalk.server.v1B\tKubeProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.kube_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\tKubeProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_KUBESERVICE"].methods_by_name["GetPodStackTraceDump"]._options = None
    _globals["_KUBESERVICE"].methods_by_name["GetPodStackTraceDump"]._serialized_options = b"\220\002\001\200}\013"
    _globals["_KUBESERVICE"].methods_by_name["GetKubernetesEvents"]._options = None
    _globals["_KUBESERVICE"].methods_by_name["GetKubernetesEvents"]._serialized_options = b"\220\002\001\200}\013"
    _globals["_KUBESERVICE"].methods_by_name["GetKubernetesPersistentVolumes"]._options = None
    _globals["_KUBESERVICE"].methods_by_name[
        "GetKubernetesPersistentVolumes"
    ]._serialized_options = b"\220\002\001\200}\013"
    _globals["_GETPODSTACKTRACEDUMPREQUEST"]._serialized_start = 159
    _globals["_GETPODSTACKTRACEDUMPREQUEST"]._serialized_end = 465
    _globals["_GETPODSTACKTRACEDUMPRESPONSE"]._serialized_start = 467
    _globals["_GETPODSTACKTRACEDUMPRESPONSE"]._serialized_end = 530
    _globals["_GETKUBERNETESEVENTSREQUEST"]._serialized_start = 533
    _globals["_GETKUBERNETESEVENTSREQUEST"]._serialized_end = 793
    _globals["_GETKUBERNETESEVENTSRESPONSE"]._serialized_start = 795
    _globals["_GETKUBERNETESEVENTSRESPONSE"]._serialized_end = 891
    _globals["_GETKUBERNETESPERSISTENTVOLUMESREQUEST"]._serialized_start = 893
    _globals["_GETKUBERNETESPERSISTENTVOLUMESREQUEST"]._serialized_end = 932
    _globals["_GETKUBERNETESPERSISTENTVOLUMESRESPONSE"]._serialized_start = 934
    _globals["_GETKUBERNETESPERSISTENTVOLUMESRESPONSE"]._serialized_end = 1054
    _globals["_KUBESERVICE"]._serialized_start = 1057
    _globals["_KUBESERVICE"]._serialized_end = 1473
# @@protoc_insertion_point(module_scope)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class RecomputeSettings(_message.Message):
    __slots__ = ("feature_fqns", "all_features")
    FEATURE_FQNS_FIELD_NUMBER: _ClassVar[int]
    ALL_FEATURES_FIELD_NUMBER: _ClassVar[int]
    feature_fqns: _containers.RepeatedScalarFieldContainer[str]
    all_features: bool
    def __init__(self, feature_fqns: _Optional[_Iterable[str]] = ..., all_features: bool = ...) -> None: ...

class CronQuery(_message.Message):
    __slots__ = (
        "name",
        "cron",
        "file_name",
        "output",
        "max_samples",
        "recompute",
        "lower_bound",
        "upper_bound",
        "tags",
        "required_resolver_tags",
        "store_online",
        "store_offline",
        "incremental_sources",
        "resource_group",
        "planner_options",
    )
    class PlannerOptionsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    NAME_FIELD_NUMBER: _ClassVar[int]
    CRON_FIELD_NUMBER: _ClassVar[int]
    FILE_NAME_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_FIELD_NUMBER: _ClassVar[int]
    MAX_SAMPLES_FIELD_NUMBER: _ClassVar[int]
    RECOMPUTE_FIELD_NUMBER: _ClassVar[int]
    LOWER_BOUND_FIELD_NUMBER: _ClassVar[int]
    UPPER_BOUND_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_RESOLVER_TAGS_FIELD_NUMBER: _ClassVar[int]
    STORE_ONLINE_FIELD_NUMBER: _ClassVar[int]
    STORE_OFFLINE_FIELD_NUMBER: _ClassVar[int]
    INCREMENTAL_SOURCES_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_GROUP_FIELD_NUMBER: _ClassVar[int]
    PLANNER_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    name: str
    cron: str
    file_name: str
    output: _containers.RepeatedScalarFieldContainer[str]
    max_samples: int
    recompute: RecomputeSettings
    lower_bound: _timestamp_pb2.Timestamp
    upper_bound: _timestamp_pb2.Timestamp
    tags: _containers.RepeatedScalarFieldContainer[str]
    required_resolver_tags: _containers.RepeatedScalarFieldContainer[str]
    store_online: bool
    store_offline: bool
    incremental_sources: _containers.RepeatedScalarFieldContainer[str]
    resource_group: str
    planner_options: _containers.ScalarMap[str, str]
    def __init__(
        self,
        name: _Optional[str] = ...,
        cron: _Optional[str] = ...,
        file_name: _Optional[str] = ...,
        output: _Optional[_Iterable[str]] = ...,
        max_samples: _Optional[int] = ...,
        recompute: _Optional[_Union[RecomputeSettings, _Mapping]] = ...,
        lower_bound: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        upper_bound: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        required_resolver_tags: _Optional[_Iterable[str]] = ...,
        store_online: bool = ...,
        store_offline: bool = ...,
        incremental_sources: _Optional[_Iterable[str]] = ...,
        resource_group: _Optional[str] = ...,
        planner_options: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/pod_request.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.chart.v1 import densetimeserieschart_pb2 as chalk_dot_chart_dot_v1_dot_densetimeserieschart__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n!chalk/server/v1/pod_request.proto\x12\x0f\x63halk.server.v1\x1a)chalk/chart/v1/densetimeserieschart.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xcc\x03\n\x1aGetPodRequestChartsRequest\x12V\n\x19start_timestamp_inclusive\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x17startTimestampInclusive\x12W\n\x17\x65nd_timestamp_exclusive\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00R\x15\x65ndTimestampExclusive\x88\x01\x01\x12>\n\rwindow_period\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0cwindowPeriod\x12\x44\n\x08grouping\x18\x04 \x01(\x0e\x32#.chalk.server.v1.PodRequestGroupingH\x01R\x08grouping\x88\x01\x01\x12N\n\x0eresource_types\x18\x05 \x03(\x0e\x32\'.chalk.server.v1.PodRequestResourceTypeR\rresourceTypesB\x1a\n\x18_end_timestamp_exclusiveB\x0b\n\t_grouping"[\n\x1bGetPodRequestChartsResponse\x12<\n\x06\x63harts\x18\x01 \x03(\x0b\x32$.chalk.chart.v1.DenseTimeSeriesChartR\x06\x63harts*\xc8\x01\n\x12PodRequestGrouping\x12$\n POD_REQUEST_GROUPING_UNSPECIFIED\x10\x00\x12"\n\x1ePOD_REQUEST_GROUPING_NAMESPACE\x10\x01\x12 \n\x1cPOD_REQUEST_GROUPING_CLUSTER\x10\x02\x12$\n POD_REQUEST_GROUPING_ENVIRONMENT\x10\x03\x12 \n\x1cPOD_REQUEST_GROUPING_SERVICE\x10\x04*\xbd\x01\n\x16PodRequestResourceType\x12)\n%POD_REQUEST_RESOURCE_TYPE_UNSPECIFIED\x10\x00\x12!\n\x1dPOD_REQUEST_RESOURCE_TYPE_CPU\x10\x01\x12$\n POD_REQUEST_RESOURCE_TYPE_MEMORY\x10\x02\x12/\n+POD_REQUEST_RESOURCE_TYPE_EPHEMERAL_STORAGE\x10\x03\x42\x98\x01\n\x13\x63om.chalk.server.v1B\x0fPodRequestProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.pod_request_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\017PodRequestProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_PODREQUESTGROUPING"]._serialized_start = 719
    _globals["_PODREQUESTGROUPING"]._serialized_end = 919
    _globals["_PODREQUESTRESOURCETYPE"]._serialized_start = 922
    _globals["_PODREQUESTRESOURCETYPE"]._serialized_end = 1111
    _globals["_GETPODREQUESTCHARTSREQUEST"]._serialized_start = 163
    _globals["_GETPODREQUESTCHARTSREQUEST"]._serialized_end = 623
    _globals["_GETPODREQUESTCHARTSRESPONSE"]._serialized_start = 625
    _globals["_GETPODREQUESTCHARTSRESPONSE"]._serialized_end = 716
# @@protoc_insertion_point(module_scope)



from chalk._gen.chalk.aggregate.v1 import service_pb2 as _service_pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.common.v1 import online_query_pb2 as _online_query_pb2
from chalk._gen.chalk.common.v1 import upload_features_pb2 as _upload_features_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class PingRequest(_message.Message):
    __slots__ = ("num",)
    NUM_FIELD_NUMBER: _ClassVar[int]
    num: int
    def __init__(self, num: _Optional[int] = ...) -> None: ...

class PingResponse(_message.Message):
    __slots__ = ("num",)
    NUM_FIELD_NUMBER: _ClassVar[int]
    num: int
    def __init__(self, num: _Optional[int] = ...) -> None: ...



from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class PointV2(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: float
    def __init__(self, value: _Optional[float] = ...) -> None: ...

class TimeSeriesV2(_message.Message):
    __slots__ = ("points", "label", "units")
    POINTS_FIELD_NUMBER: _ClassVar[int]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    UNITS_FIELD_NUMBER: _ClassVar[int]
    points: _containers.RepeatedCompositeFieldContainer[PointV2]
    label: str
    units: str
    def __init__(
        self,
        points: _Optional[_Iterable[_Union[PointV2, _Mapping]]] = ...,
        label: _Optional[str] = ...,
        units: _Optional[str] = ...,
    ) -> None: ...

class TimeSeriesChartV2(_message.Message):
    __slots__ = ("title", "series", "x_series", "window_period")
    TITLE_FIELD_NUMBER: _ClassVar[int]
    SERIES_FIELD_NUMBER: _ClassVar[int]
    X_SERIES_FIELD_NUMBER: _ClassVar[int]
    WINDOW_PERIOD_FIELD_NUMBER: _ClassVar[int]
    title: str
    series: _containers.RepeatedCompositeFieldContainer[TimeSeriesV2]
    x_series: _containers.RepeatedCompositeFieldContainer[_timestamp_pb2.Timestamp]
    window_period: _duration_pb2.Duration
    def __init__(
        self,
        title: _Optional[str] = ...,
        series: _Optional[_Iterable[_Union[TimeSeriesV2, _Mapping]]] = ...,
        x_series: _Optional[_Iterable[_Union[_timestamp_pb2.Timestamp, _Mapping]]] = ...,
        window_period: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
    ) -> None: ...



from __future__ import annotations

import dataclasses
import time
from typing import TYPE_CHECKING, Callable, Literal, Sequence, TypeVar, final

import grpc

from chalk._gen.chalk.server.v1.auth_pb2 import GetTokenRequest, GetTokenResponse
from chalk._gen.chalk.server.v1.auth_pb2_grpc import AuthServiceStub

if TYPE_CHECKING:
    from chalk import EnvironmentId


@dataclasses.dataclass
class _ClientCallDetails(grpc.ClientCallDetails):
    method: str
    timeout: float | None
    metadata: grpc.Metadata | None
    credentials: grpc.CallCredentials | None


@final
class TokenRefresher:
    def __init__(
        self,
        auth_stub: AuthServiceStub,
        client_id: str,
        client_secret: str,
    ):
        self._auth_stub = auth_stub
        self._client_id = client_id
        self._client_secret = client_secret
        self._auth_token: GetTokenResponse | None = None

    def get_token(self) -> GetTokenResponse:
        if self._auth_token is None or self._auth_token.expires_at.seconds - time.time() <= 60:
            self._auth_token = self._auth_stub.GetToken(
                GetTokenRequest(
                    client_id=self._client_id,
                    client_secret=self._client_secret,
                    grant_type="client_credentials",
                ),
            )

        return self._auth_token


RequestType = TypeVar("RequestType")
ResponseType = TypeVar("ResponseType")


@final
class AuthenticatedChalkClientInterceptor(grpc.UnaryUnaryClientInterceptor):
    """
    This GRPC Client Interceptor, adds an auth token and default
    Chalk headers to a grpc channel.
    """

    def __init__(
        self,
        refresher: TokenRefresher,
        environment_id: EnvironmentId | None,
        server: Literal["go-api", "engine"],
        additional_headers: list[tuple[str, str]],
    ):
        self._refresher: TokenRefresher = refresher
        self._constant_headers = [
            ("x-chalk-server", server),
            *additional_headers,
        ]
        if environment_id is not None:
            self._constant_headers.append(("x-chalk-env-id", environment_id))

    def intercept_unary_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, RequestType], grpc.CallFuture[ResponseType]],
        client_call_details: grpc.ClientCallDetails,
        request: RequestType,
    ) -> grpc.CallFuture[ResponseType]:
        headers: dict[str, str | bytes] = dict(self._constant_headers)
        headers["authorization"] = f"Bearer {self._refresher.get_token().access_token}"
        if client_call_details.metadata:
            headers.update(client_call_details.metadata)
        return continuation(
            _ClientCallDetails(
                method=client_call_details.method,
                timeout=client_call_details.timeout,
                metadata=tuple(headers.items()),
                credentials=client_call_details.credentials,
            ),
            request,
        )


@final
class UnauthenticatedChalkClientInterceptor(grpc.UnaryUnaryClientInterceptor):
    """
    This GRPC Client Interceptor, adds an auth token and default
    Chalk headers to a grpc channel.
    """

    def __init__(
        self,
        additional_headers: Sequence[tuple[str, str]],
        server: Literal["go-api", "engine"],
    ):
        self._headers = (
            ("x-chalk-server", server),
            *additional_headers,
        )

    def intercept_unary_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, RequestType], grpc.CallFuture[ResponseType]],
        client_call_details: grpc.ClientCallDetails,
        request: RequestType,
    ) -> grpc.CallFuture[ResponseType]:
        if client_call_details.metadata is None:
            headers = self._headers
        else:
            headers_dict: dict[str, str | bytes] = dict(self._headers)
            headers_dict.update(client_call_details.metadata)
            headers = tuple(headers_dict.items())
        return continuation(
            _ClientCallDetails(
                method=client_call_details.method,
                timeout=client_call_details.timeout,
                metadata=headers,
                credentials=client_call_details.credentials,
            ),
            request,
        )



from __future__ import annotations

import io
import math
import pathlib
import typing
from io import BytesIO
from typing import TYPE_CHECKING, Any, BinaryIO, Literal, Union

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.ipc as pa_ipc
import pyarrow.parquet as pq
import requests

from chalk.features._encoding.pyarrow import is_map_in_dtype_tree, pyarrow_to_polars
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import polars as pl


_logger = get_logger(__name__)

T_ArrowTable = typing.TypeVar("T_ArrowTable", bound=typing.Union[pa.RecordBatch, pa.Table])


class ArrowTableCastError(Exception):
    def __init__(self, errors: list[ArrowColumnCastError]):
        self.errors = errors
        super().__init__(f"Failed to cast columns: {errors}")


class ArrowColumnCastError(Exception):
    def __init__(self, msg: str, col_name: str, expected_dtype: pa.DataType, actual_dtype: pa.DataType):
        self.col_name = col_name
        self.expected_dtype = expected_dtype
        self.actual_dtype = actual_dtype
        super().__init__(
            f"Failed to cast column '{col_name}' from '{actual_dtype}' to '{expected_dtype}': {msg}",
        )


def is_binary_like(dtype: pa.DataType):
    return pa.types.is_binary(dtype) or pa.types.is_large_binary(dtype) or pa.types.is_fixed_size_binary(dtype)


def is_list_like(dtype: pa.DataType):
    return pa.types.is_list(dtype) or pa.types.is_large_list(dtype) or pa.types.is_fixed_size_list(dtype)


class PyArrowToPolarsConverter:
    """Convert a pyarrow table to a polars dataframe. Unlike ``pl.from_arrow``, this function correctly
    casts the schema where polars might otherwise choose the incorrect data type.

    This class requires the schema to be provided separately so the pyarrow-to-polars conversion expression
    can be cached
    """

    def __init__(self, schema: pa.Schema, use_fixed_size_list: bool = False) -> None:
        super().__init__()
        self.schema = schema
        self.no_ext_schema = self._remove_extension_types(schema)
        self._pl_schema = {
            col_name: pyarrow_to_polars(schema.field(col_name).type, col_name, use_fixed_size_list=use_fixed_size_list)
            for col_name in schema.names
        }

    def _does_dtype_have_float16(self, dtype: pa.DataType) -> bool:
        if pa.types.is_float16(dtype):
            return True
        for fidx in range(dtype.num_fields):
            if self._does_dtype_have_float16(dtype.field(fidx).type):
                return True
        return False

    def _recursive_convert_map(self, arr: Union[pa.Array, pa.ChunkedArray]) -> pa.Array:
        """
        Recursively converts a `MapArray` to a `LargeList` of `StructArray`s,
        because `pl.from_arrow` cannot handle nested maps.
        (
            fails with:
               pyo3_runtime.PanicException: Arrow datatype Map(...) not supported by Polars.
               You probably need to activate that data-type feature.
        )
        """
        if isinstance(arr, pa.ChunkedArray):
            arr = arr.combine_chunks()

        if isinstance(arr, pa.MapArray):
            """
            ==================
            Original map array
            ==================

                MapArray
                    [
                        [
                           {key: abc, value: bbc},
                           {key: efg, value: hij},
                        ],
                        [
                           {key: ddd, value: bbc},
                           {key: zzz, value: hij},
                        ]
                    ]


                keys = [abc, efg, ddd, zzz]
                items = [bbc, hij, bbc, hij]
                offsets = [0, 2, 4]

            =========================
            Intermediate struct array
            =========================

                [
                    [
                       {key: abc, value: bbc},
                       {key: efg, value: hij},
                    ],
                    [
                       {key: ddd, value: bbc},
                       {key: zzz, value: hij},
                    ]
                ]

            ----------------------------------------
            To create from intermediate struct array:
            ----------------------------------------

                pa.ListArray.from_arrays([0, 2, 4], struct_array)

            """
            struct_arr = pa.StructArray.from_arrays(
                [arr.keys, self._recursive_convert_map(arr.items)], ["key", "value"]
            )
            list_arr = pa.LargeListArray.from_arrays(arr.offsets, struct_arr, mask=arr.is_null())
            return list_arr
        elif isinstance(arr, (pa.LargeListArray, pa.ListArray)):
            return pa.LargeListArray.from_arrays(
                arr.offsets, self._recursive_convert_map(arr.values), mask=arr.is_null()
            )
        elif isinstance(arr, pa.FixedSizeListArray):
            return pa.FixedSizeListArray.from_arrays(self._recursive_convert_map(arr.values), arr.type.list_size)
        elif isinstance(arr, pa.StructArray):
            return pa.StructArray.from_arrays(
                [self._recursive_convert_map(arr.field(i)) for i in range(arr.type.num_fields)],
                names=[arr.type[i].name for i in range(arr.type.num_fields)],
            )
        else:
            return arr

    @classmethod
    def _recursive_convert_float16(cls, x: pa.Array) -> pa.Array:
        """
        Recursively replace float16 columns with float32 since Polars doesn't support float16.
        1. If col is float16, cast to float32
        2. If col is a fixed-size list of float16's, convert the underlying values to float32 and fill in nulls w/ lists of null values of the apprp length.
        :param x:
        :return:
        """
        if pa.types.is_float16(x.type):
            return pa.array(x.to_numpy(zero_copy_only=False).astype(np.dtype("float32")))
        if x.type.num_fields == 0:
            # if it's not a nested type and it's not float16, return as is
            return x

        if pa.types.is_fixed_size_list(x.type):
            assert isinstance(x.type, pa.FixedSizeListType)
            if pa.types.is_float16(x.type.value_type):
                assert isinstance(x, pa.FixedSizeListArray)
                # We'll first expand all null elements into null lists of the correct length, and then convert to numpy
                null_elements = x.is_null()
                empty = pa.scalar(
                    np.zeros((x.type.list_size,), dtype=np.dtype("float16")),
                    pa.list_(pa.float16(), x.type.list_size),
                )
                x = x.fill_null(empty)
                pa_arr = pa.FixedSizeListArray.from_arrays(
                    x.flatten().to_numpy(zero_copy_only=False).astype(np.dtype("float32")),
                    x.type.list_size,
                )
                # Replace the filled empty elements with null
                pa_arr = pc.if_else(  # type: ignore
                    null_elements, pa.scalar(None, pa.list_(pa.float32(), x.type.list_size)), pa_arr
                )
                return pa_arr
            # TODO: nested fixed-size-lists e.g. FixedSizeList<FixedSizeList<float16>>
        if pa.types.is_struct(x.type):
            assert isinstance(x, pa.StructArray)
            child_arrays = x.flatten()
            child_arrays = [cls._recursive_convert_float16(x) for x in child_arrays]
            child_names = [x.name for x in x.type.fields]
            converted_struct_array = pa.StructArray.from_arrays(
                arrays=child_arrays, names=child_names, mask=x.is_null()
            )
            # Apply the original null mask to new struct array
            return pc.if_else(pc.is_null(x), None, converted_struct_array)  # type: ignore
        if pa.types.is_list(x.type) or pa.types.is_large_list(x.type):
            values_converted = x.flatten()
            values_converted = cls._recursive_convert_float16(values_converted)
            # (call class method w/ the object to get the correct sub-class)
            return x.from_arrays(offsets=x.offsets, values=values_converted, mask=x.is_null())
        return x

    def convert(self, table: pa.Table) -> pl.DataFrame:
        import polars as pl

        assert table.schema == self.schema, "The table schema differs from the declared schema"

        table = table.cast(self.no_ext_schema)

        new_cols: list[pa.Array | pa.ChunkedArray] = []
        for x in table.columns:
            if self._does_dtype_have_float16(x.type):
                x = self._recursive_convert_float16(x.combine_chunks())
            if is_map_in_dtype_tree(x.type):
                new_cols.append(self._recursive_convert_map(x))
                continue
            new_cols.append(x)
        table = pa.Table.from_arrays(new_cols, table.column_names)

        try:
            df = pl.from_arrow(table, self._pl_schema)
        except Exception:
            _logger.debug(
                f"pl.from_arrow failed. Trying again after combining chunks and viewing, {table.num_rows=}, {table.nbytes=}, {table.schema=}",
                exc_info=True,
            )
            # Sometimes the table will have null buffers, which polars cannot handle. But it will work if we view it as itself
            # why? who knows
            table = pa.Table.from_pydict(
                {k: v.combine_chunks().view(v.type) for (k, v) in zip(table.column_names, table.columns)}
            )
            try:
                df = pl.from_arrow(table, self._pl_schema)
            except Exception:
                _logger.debug(
                    f"Trying to deal with table without chunking, {table.num_rows=}, {table.nbytes=}, {table.schema=}",
                    exc_info=True,
                )
                df = pl.from_arrow(table, rechunk=False, schema=self._pl_schema)

        assert isinstance(df, pl.DataFrame)
        col_name_to_expr = {
            col_name: pl.col(col_name).cast(expected_dtype).alias(col_name)
            for (col_name, actual_dtype) in df.schema.items()
            if (expected_dtype := self._pl_schema[col_name]) != actual_dtype
        }
        if len(col_name_to_expr) > 0:
            _logger.warning(f"PyArrow <-> polars schema mismatch for columns {', '.join(col_name_to_expr.keys())}")
            df = df.with_columns(list(col_name_to_expr.values()))
        return df

    @staticmethod
    def _source_type_of_extension_type(dtype: pa.DataType) -> pa.DataType:
        if isinstance(dtype, pa.ExtensionType):
            return dtype.storage_type
        return dtype

    @staticmethod
    def _remove_extension_types(table: pa.Schema) -> pa.Schema:
        return pa.schema(
            [
                pa.field(
                    name=field.name,
                    type=PyArrowToPolarsConverter._source_type_of_extension_type(field.type),
                    metadata=field.metadata,
                )
                for field in table
            ]
        )


def pa_table_to_pl_df(table: pa.Table[Any]) -> pl.DataFrame:
    return PyArrowToPolarsConverter(table.schema).convert(table)


def pa_array_to_pl_series(arr: pa.Array | pa.ChunkedArray) -> pl.Series:
    tbl = pa.Table.from_arrays([arr], ["col_0"])
    df = pa_table_to_pl_df(tbl)
    return df.get_column("col_0")


def pa_cast(table: pa.Table[Any], expected_schema: pa.Schema, collect_all_errors: bool = False) -> pa.Table[Any]:
    """Safely cast a pyarrow table to the expected schema. Unlike ``table.cast(schema)``, this function will reorder struct columns if needed"""
    table_column_names = list(table.column_names)
    table_col_name_to_col = {col_name: col for (col_name, col) in zip(table_column_names, table.columns)}
    expected_column_names = list(expected_schema.names)
    assert frozenset(expected_column_names).issubset(
        table_column_names
    ), f"The expected column names ({expected_column_names}) must be a subset of the table column names ({table_column_names})."
    if table.schema == expected_schema:
        # Short circuit
        return table
    # First let's select just the columns we're interested in
    table = table.select(expected_column_names)

    if table.schema == expected_schema:
        # Short circuit
        return table

    new_arrays: list[pa.ChunkedArray | pa.Array] = []

    errors: list[ArrowColumnCastError] = []
    for name, expected_type in zip(expected_column_names, expected_schema.types):
        col = table_col_name_to_col[name]
        assert isinstance(col, pa.ChunkedArray)
        if len(col) == 0:
            arr = pa.array([], expected_type)
            new_arrays.append(arr)
            continue

        casted_chunks = []
        for chunk in col.chunks:
            if len(chunk) == 0:
                continue
            try:
                chunk_res = pa_cast_col(chunk, expected_type)
            except Exception as e:
                wrapped_error = ArrowColumnCastError(
                    msg=str(e), col_name=name, expected_dtype=expected_type, actual_dtype=chunk.type
                )
                if not collect_all_errors:
                    raise wrapped_error from e
                errors.append(wrapped_error)
            else:
                casted_chunks.append(chunk_res)
        if len(casted_chunks) == 0:
            arr = pa.nulls(0, expected_type)
        else:
            arr = pa.chunked_array(casted_chunks)
        new_arrays.append(arr)
    if errors:
        raise ArrowTableCastError(errors)

    return pa.Table.from_arrays(new_arrays, names=expected_schema.names)


def recursive_convert_map_primitive(x: Any, dtype: pa.DataType):
    """
    When we call polars `to_arrow` we end up with a DataFrame that
    has a bunch of maps or nested maps that are indistinguishable
    from list of structs of `key` and `value` fields. This function
    converts those list of dicts with `key` `value` fields into a dict,
    and they can then be faithfully converted into their `rich` form.
    """
    if x is None:
        return x
    if isinstance(dtype, pa.MapType):
        if not isinstance(x, list):
            raise ValueError(f"Expected a list, but got {type(x).__name__}")
        return {e[0]: recursive_convert_map_primitive(e[1], dtype.item_type) for e in x}
    elif isinstance(dtype, (pa.ListType, pa.LargeListType, pa.FixedSizeListType)):
        if not isinstance(x, list):
            raise ValueError(f"Expected a list, but got {type(x).__name__}")
        return [recursive_convert_map_primitive(y, dtype.value_type) for y in x]
    elif isinstance(dtype, pa.StructType):
        if not isinstance(x, dict):
            raise ValueError(f"Expected a dict, but got {type(x).__name__}")
        res = {}
        for k, v in x.items():
            field_idx = dtype.get_field_index(k)
            if field_idx == -1:
                raise ValueError(f"Missing field '{k}' in dtype '{dtype}'")
            res[k] = recursive_convert_map_primitive(v, dtype.field(field_idx).type)
        return res
    return x


def _recursive_convert_map_type(dtype: pa.DataType) -> pa.DataType:
    if pa.types.is_map(dtype):
        assert isinstance(dtype, pa.MapType)
        return pa.large_list(
            pa.struct(
                [
                    pa.field(name="key", type=dtype.key_type),
                    pa.field(name="value", type=_recursive_convert_map_type(dtype.item_type)),
                ]
            )
        )
    elif pa.types.is_large_list(dtype):
        assert isinstance(dtype, pa.LargeListType)
        return pa.large_list(_recursive_convert_map_type(dtype.value_type))
    elif pa.types.is_list(dtype):
        assert isinstance(dtype, pa.ListType)
        return pa.list_(_recursive_convert_map_type(dtype.value_type))
    elif pa.types.is_fixed_size_list(dtype):
        assert isinstance(dtype, pa.FixedSizeListType)
        return pa.list_(_recursive_convert_map_type(dtype.value_type), dtype.list_size)
    elif pa.types.is_struct(dtype):
        assert isinstance(dtype, pa.StructType)
        return pa.struct(
            [
                pa.field(
                    name=field.name,
                    type=_recursive_convert_map_type(field.type),
                    nullable=field.nullable,
                    metadata=field.metadata,
                )
                for field in dtype
            ]
        )
    return dtype


def pa_cast_col(col: pa.Array, expected_type: pa.DataType) -> pa.Array:
    """
    Converts the provided input type to the provided expected type.
    This conversion is used to avoid hard-to-represent types, like fixed-sized-lists or maps, which
    polars in particular does not support.
    """
    if col.null_count == len(col):
        # It's all null, so short circuit and return an array of nulls of the correct type
        # calling .flatten() sometimes will raise an exception if this condition is true
        return pa.nulls(len(col), expected_type)
    if col.type == expected_type:
        return col
    if pa.types.is_map(expected_type):
        assert isinstance(expected_type, pa.MapType)
        # Validate that the arrow "maps" are in list<struct<key,value>> format
        if not (
            (isinstance(col, pa.LargeListArray) or isinstance(col, pa.ListArray))
            and hasattr(col, "values")
            and isinstance(col.values, pa.StructArray)
            and isinstance(struct := col.values.type, pa.StructType)
            and struct.get_field_index("key") == 0
            and struct.get_field_index("value") == 1
        ):
            raise ValueError(f"Expected the column to be a list of <key,value> structs, instead got {col.type}")

        keys = pa_cast_col(col.values.field("key"), expected_type.key_type)
        values = pa_cast_col(col.values.field("value"), expected_type.item_type)
        return pa.MapArray.from_arrays(col.offsets, keys, values, mask=col.is_null())
    if pa.types.is_struct(expected_type):
        # Convert the column to a table, then recursively cast the table
        assert isinstance(expected_type, pa.StructType)
        assert isinstance(col, pa.StructArray)
        arrays: list[pa.Array] = []
        names: list[str] = []
        fields: list[pa.Field] = []
        struct_type_as_schema = pa.schema(expected_type)
        for i in range(len(struct_type_as_schema)):
            field = struct_type_as_schema.field(i)
            fields.append(field)
            arrays.append(col.field(field.name))
            names.append(field.name)
        expected_sub_schema = pa.schema(fields)
        tbl = pa.Table.from_arrays(arrays, names)
        tbl = pa_cast(tbl, expected_sub_schema)
        # Now convert it back into a struct
        casted_columns = [x.combine_chunks() if isinstance(x, pa.ChunkedArray) else x for x in tbl.columns]
        struct_array = pa.StructArray.from_arrays(casted_columns, tbl.column_names)
        struct_array = struct_array.cast(expected_type)  # should be a no-op?
        return struct_array
    if pa.types.is_list(expected_type) or pa.types.is_large_list(expected_type):
        assert isinstance(expected_type, (pa.LargeListType, pa.ListType))
        assert isinstance(col, (pa.ListArray, pa.LargeListArray, pa.FixedSizeListArray))
        flattened = col.flatten()
        tbl = pa.Table.from_arrays([flattened], names=[expected_type.value_field.name])
        tbl = pa_cast(tbl, pa.schema([expected_type.value_field]))
        casted_col = tbl.column(0).combine_chunks()
        assert len(flattened) == len(
            casted_col
        ), f"flattened has length {len(flattened)} but casted has length {len(casted_col)}"
        # col.offsets has random-looking values if the array has nulls. So we'll calculate the offsets
        # ourselves
        arr = col.value_lengths().fill_null(0)
        offsets: pa.Array = pc.cumulative_sum_checked(arr)  # type: ignore
        zero = pa.array([0], offsets.type)
        assert isinstance(zero, pa.Array)
        offsets = pa.concat_arrays([zero, offsets])
        mask = col.is_valid()
        single_true = pa.array([True], type=pa.bool_())
        assert isinstance(single_true, pa.Array)
        mask = pa.concat_arrays([mask, single_true])
        assert len(mask) == len(offsets)

        # Select the offset if the mask is True. Otherwise replace with null
        # This is per the docstring of ListArray.from_arrays
        # https://arrow.apache.org/docs/python/generated/pyarrow.ListArray.html#pyarrow.ListArray.from_arrays
        nulls = pa.nulls(len(offsets), offsets.type)
        offsets = pc.if_else(mask, offsets, nulls)  # type: ignore

        if pa.types.is_list(expected_type):
            if not isinstance(offsets, pa.Int32Array):
                offsets = offsets.cast(pa.int32())
            assert isinstance(offsets, pa.Int32Array)

            ans = pa.ListArray.from_arrays(offsets, casted_col)
        else:
            assert pa.types.is_large_list(expected_type)
            if not isinstance(offsets, pa.Int64Array):
                offsets = offsets.cast(pa.int64())
            assert isinstance(offsets, pa.Int64Array)
            ans = pa.LargeListArray.from_arrays(offsets, casted_col)

        assert len(ans) == len(col), "array should have the same number of elements"
        return ans
    if pa.types.is_fixed_size_list(expected_type):
        # Cast the elements of the input column, and then reconstruct an outer list.
        assert isinstance(col, (pa.ListArray, pa.LargeListArray, pa.FixedSizeListArray))
        if isinstance(col, pa.ListArray):
            col_items_converted = pa_cast_col(col.values, expected_type=expected_type.value_type)
            rebuilt_col = pa.ListArray.from_arrays(
                offsets=col.offsets,
                values=col_items_converted,
                type=pa.list_(expected_type.value_type),
            )
            if col.null_count != 0:
                # Pyarrow does not allow `mask` to be set if `offset` is non-zero.
                # To support this case, null out the missing lists with `if_else`.

                # Note: we are missing type stubs for `pc.if_else`
                rebuilt_col = pc.if_else(  # pyright: ignore[reportAttributeAccessIssue]
                    col.is_null(),
                    pa.nulls(len(rebuilt_col), rebuilt_col.type),
                    rebuilt_col,
                )
        elif isinstance(col, pa.LargeListArray):
            col_items_converted = pa_cast_col(col.values, expected_type=expected_type.value_type)
            rebuilt_col = pa.LargeListArray.from_arrays(
                offsets=col.offsets,
                values=col_items_converted,
                type=pa.large_list(expected_type.value_type),
            )
            if col.null_count != 0:
                # Pyarrow does not allow `mask` to be set if `offset` is non-zero.
                # To support this case, null out the missing lists with `if_else`.

                # Note: we are missing type stubs for `pc.if_else`
                rebuilt_col = pc.if_else(  # pyright: ignore[reportAttributeAccessIssue]
                    col.is_null(),
                    pa.nulls(len(rebuilt_col), rebuilt_col.type),
                    rebuilt_col,
                )
        elif isinstance(col, pa.FixedSizeListArray):
            col_items_converted = pa_cast_col(col.values, expected_type=expected_type.value_type)
            rebuilt_col = pa.FixedSizeListArray.from_arrays(
                values=col_items_converted,
                type=pa.list_(expected_type.value_type, col.type.list_size),
            )
            if col.null_count != 0:
                # Pyarrow does not allow `mask` to be set if `offset` is non-zero.
                # To support this case, null out the missing lists with `if_else`.

                # Note: we are missing type stubs for `pc.if_else`
                rebuilt_col = pc.if_else(  # pyright: ignore[reportAttributeAccessIssue]
                    col.is_null(),
                    pa.nulls(len(rebuilt_col), rebuilt_col.type),
                    rebuilt_col,
                )
        else:
            raise ValueError(
                f"Cannot convert array with non-list type {col.type} to fixed-sized list type '{expected_type}'"
            )

        return pc.cast(rebuilt_col, expected_type)
    if pa.types.is_map(expected_type):
        return pa_cast_col(col, expected_type)

    # Otherwise, cast directly if it's a scalar

    try:
        return col.cast(expected_type)
    except Exception as e:
        if pa.types.is_large_string(col.type) and pa.types.is_string(expected_type):
            _logger.error(
                f"Casting large string to large string instead of string like a maniac, {col.nbytes=}, {len(col)=}"
            )
            return col

        raise e


def pa_replace_column(t: T_ArrowTable, col_name: str, col: pa.Array | pa.ChunkedArray) -> T_ArrowTable:
    idx = t.schema.get_field_index(col_name)
    if idx < 0:
        raise KeyError(f"Could not find unique column with name {col_name} in table.")
    return t.set_column(idx, col_name, col)


def _read_parquet(
    source: str | pathlib.Path | BinaryIO | BytesIO | bytes,
    **kwargs: Any,
):
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")

    if isinstance(source, str):
        if source.startswith("file://"):
            # Polars cannot handle local file uris properly
            source = source[len("file://") :]
        if source.startswith("gs://"):
            # Fsspec expects gcs instead of gs
            source = "gcs://" + source[len("gs://") :]

    try:
        # Polars added a crash bug in polars 0.20 that isn't fixed until polars 1.8.0; this fallback case handles
        # that issue with a less efficient alternative that avoids their bug reading signed uris.
        return pl.read_parquet(source, **kwargs)
    except:
        if isinstance(source, str):
            if source.startswith("https://"):
                buffer = BytesIO(requests.get(source).content)
            else:
                buffer = open(source, "rb")
            try:
                return pl.read_parquet(buffer, **kwargs)
            except:
                buffer.seek(0)
                import pyarrow.parquet

                pa_table = pyarrow.parquet.read_table(buffer, **kwargs)
                pl_table = PyArrowToPolarsConverter(schema=pa_table.schema, use_fixed_size_list=True).convert(pa_table)
                return pl_table
        else:
            raise


if TYPE_CHECKING:
    read_parquet = pl.read_parquet
else:
    read_parquet = _read_parquet


def record_batch_to_arrow_ipc(rb: pa.RecordBatch, compression: Literal["lz4", "zstd", "uncompressed"] = "lz4"):
    dest = BytesIO()
    writer = pa_ipc.RecordBatchFileWriter(dest, rb.schema, options=pa_ipc.IpcWriteOptions(compression=compression))
    writer.write_batch(rb)
    writer.close()
    dest.seek(0)
    feather_bytes = dest.read()
    return feather_bytes


def arrow_ipc_to_record_batch(b: bytes) -> pa.RecordBatch:
    bio = io.BytesIO(b)
    reader = pa_ipc.RecordBatchFileReader(bio)
    t: pa.Table = reader.read_all()
    return pa_table_to_recordbatch(t)


def pa_table_to_recordbatch(t: pa.Table) -> pa.RecordBatch:
    return pa.record_batch([c.combine_chunks() for c in t.columns], t.column_names)


def estimate_row_size_bytes(
    dtype: pa.DataType,
    *,
    average_string_size: int = 30,
    average_bytes_size: int = 100,
    average_list_length: int = 5,
    null_fraction: float = 0.0,
    float16_size: int = 2,
) -> float:
    """Estimate the average size of a row, given a schema. This is only slightly better than hardcoding logic based on row count, since it takes into account the width of the table.
    However, if the estimates for string, bytes, or list sizes are off, then this estimate could be way off. It assumes no nulls"""
    # Adding 0.125 to most fields to account for the nil bit
    if pa.types.is_null(dtype):
        return 0.125  # just the nil bit
    if pa.types.is_boolean(dtype):
        return 0.250  # bool bit + nil bit
    if pa.types.is_int8(dtype) or pa.types.is_uint8(dtype):
        return 1.125
    if pa.types.is_int16(dtype) or pa.types.is_uint16(dtype):
        return 2.125
    if (
        pa.types.is_int32(dtype)
        or pa.types.is_uint32(dtype)
        or pa.types.is_date32(dtype)
        or pa.types.is_time32(dtype)
        or pa.types.is_float32(dtype)
    ):
        return 4.125
    if (
        pa.types.is_int64(dtype)
        or pa.types.is_uint64(dtype)
        or pa.types.is_date64(dtype)
        or pa.types.is_time64(dtype)
        or pa.types.is_float64(dtype)
        or pa.types.is_timestamp(dtype)
        or pa.types.is_duration(dtype)
    ):
        return 8.125
    if pa.types.is_float16(dtype):
        return float16_size + 0.125
    if pa.types.is_string(dtype) or pa.types.is_large_string(dtype):
        return average_string_size + 0.125
    if pa.types.is_binary(dtype) or pa.types.is_large_binary(dtype):
        return average_bytes_size + 0.125
    if pa.types.is_fixed_size_binary(dtype):
        assert isinstance(dtype, pa.FixedSizeBinaryType)
        return dtype.byte_width + 0.125
    if pa.types.is_list(dtype) or pa.types.is_large_list(dtype) or pa.types.is_fixed_size_list(dtype):
        assert isinstance(dtype, (pa.ListType, pa.LargeListType, pa.FixedSizeListType))
        return (dtype.list_size if isinstance(dtype, pa.FixedSizeListType) else average_list_length) * (
            estimate_row_size_bytes(
                dtype.value_type,
                average_string_size=average_string_size,
                average_bytes_size=average_bytes_size,
                average_list_length=average_list_length,
                null_fraction=null_fraction,
                float16_size=float16_size,
            )
        ) + 0.125
    if pa.types.is_map(dtype):
        assert isinstance(dtype, pa.MapType)
        key_type = dtype.key_type
        value_type = dtype.item_type
        return (
            estimate_row_size_bytes(
                key_type,
                average_string_size=average_string_size,
                average_bytes_size=average_bytes_size,
                average_list_length=average_list_length,
                null_fraction=null_fraction,
                float16_size=float16_size,
            )
            + estimate_row_size_bytes(
                value_type,
                average_string_size=average_string_size,
                average_bytes_size=average_bytes_size,
                average_list_length=average_list_length,
                null_fraction=null_fraction,
                float16_size=float16_size,
            )
        ) * average_list_length + 0.125
    if pa.types.is_struct(dtype):
        assert isinstance(dtype, pa.StructType)
        size = 0.125
        for i in range(dtype.num_fields):
            f = dtype.field(i)
            size += estimate_row_size_bytes(
                f.type,
                average_string_size=average_string_size,
                average_bytes_size=average_bytes_size,
                average_list_length=average_list_length,
                null_fraction=null_fraction,
                float16_size=float16_size,
            )
        return size

    # If there's a type we didn't handle, assume 4 bytes
    return 4.125


def write_table_to_parquet(table: pa.Table, compression: str = "snappy") -> io.BytesIO:
    buffer = io.BytesIO()
    # From https://cloud.google.com/bigquery/docs/loading-data-cloud-storage-parquet#input_file_requirements,
    # row sizes need to be less than 50MB. I presume they meant to say row GROUP sizes in the docs, not individual row
    # sizes. Assuming each row is approximately the same size, we can determine an appropriate row group size
    if len(table) == 0:
        # Doesn't really matter since there is no data!
        target_row_group_size = 1
    else:
        avg_bytes_per_row = table.nbytes / len(table)
        max_bytes_per_row_group = 50 * 1024 * 1024
        avg_row_group_size = max_bytes_per_row_group / avg_bytes_per_row
        # FIXME: Pick the target row group size via statistics. But I think targeting for 25MB row groups (i.e. half of the allowance)
        # should be pretty safe
        target_row_group_size = int(avg_row_group_size // 2)
        # The default row group size is 1024*1024 (https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_table.html#pyarrow.parquet.write_table)
        # Let's never exceed that
        target_row_group_size = min(1024 * 1024, target_row_group_size)
        # But let's always have at least one row
        target_row_group_size = max(target_row_group_size, 1)
    pq.write_table(
        table,
        buffer,
        compression=compression,
        use_compliant_nested_type=True,
        row_group_size=target_row_group_size,
    )
    buffer.seek(0)
    return buffer


def describe_type_of_dataframe(frame: pl.DataFrame | pa.Table) -> str:
    description = f"DataFrame shape=({frame.shape[0]} rows, {frame.shape[1]} cols)"

    if isinstance(frame, pa.Table):
        frame_columns = frame.column_names  # pyright: ignore[reportAttributeAccessIssue]
    else:
        frame_columns = frame.columns

    if len(frame_columns) > 0:
        if len(frame_columns) < 20:
            description += f" with columns {repr(frame_columns)}"
        else:
            description += f" with selected columns {repr(frame_columns[:20])}"

    return description


def to_pydict(arrow_obj: pa.RecordBatch | pa.Table, /):
    """Convert an arrow object to a python list using our optimized impl"""
    return dict(zip(arrow_obj.column_names, (to_pylist(x) for x in arrow_obj.columns)))


def to_pylist(arrow_obj: pa.Array | pa.ChunkedArray, /) -> list[object]:
    """Convert an arrow object to a python list using our optimized impl"""
    try:
        from libchalk.utils import (  # pyright: ignore[reportMissingModuleSource,reportMissingImports]
            arrow_array_to_pylist,
        )
    except ImportError:
        return arrow_obj.to_pylist()
    else:
        if isinstance(arrow_obj, pa.ChunkedArray):
            ans: list[object] = []
            for chunk in arrow_obj.chunks:
                if isinstance(chunk, pa.ExtensionArray):
                    assert isinstance(
                        arrow_obj.type, pa.ExtensionType
                    ), "ExtensionArray must have an ExtensionType type."
                    chunk = chunk.storage.cast(arrow_obj.type.storage_type)
                ans.extend(arrow_array_to_pylist(chunk))
            return ans
        elif isinstance(arrow_obj, pa.ExtensionArray):
            assert isinstance(arrow_obj.type, pa.ExtensionType), "ExtensionArray must have an ExtensionType type."
            arrow_obj = arrow_obj.storage.cast(arrow_obj.type.storage_type)
        return arrow_array_to_pylist(arrow_obj)


def chunk_table(table: pa.Table, n: int) -> list:
    """
    Split a PyArrow Table into exactly `n` chunks.

    Parameters:
        table (pa.Table): The PyArrow Table to chunk.
        n (int): The number of chunks to create.

    Returns:
        list: A list of PyArrow Tables, each forming a chunk.
    """
    if n <= 0:
        raise ValueError("Number of chunks `n` must be greater than 0.")

    if n == 1:
        return [table]  # Return the whole table as a single chunk

    total_rows = table.num_rows

    if n >= total_rows:
        # If n is greater than or equal to total rows, each chunk will have one row
        return [table.slice(i, 1) for i in range(total_rows)]

    chunk_size = math.ceil(total_rows / n)
    chunks = [table.slice(i, chunk_size) for i in range(0, total_rows, chunk_size)]

    return chunks
